unit ubancodebaterias01;

interface

uses
  Classes,
  xMatDefs, uSimplex, ucosa, uGlobs, uFechas,
  ugeneradores,
  uNodos,
  SysUtils,
  uEstados,
  uFichasLPD,
  uconstantesSimSEE,
  uevapUruguay;

resourcestring
  rsBancoDeBaterias01 = 'Banco de Baterías';

type
  TBancoDeBaterias01 = class; // forward declaration



  TFichaBancoDeBaterias01 = class(TFichaLPD)
  public

    PMax_Gen: NReal;
    PMax_Dem: NReal;
    ren_Gen: NReal;
    ren_Dem: NReal;
    CapacidadMaxima_MWh: NReal;
    fd: NReal;
    tmr: NReal;



    constructor Create(capa: integer; fecha: TFecha; periodicidad: TPeriodicidad;
      PMax_Gen, PMax_Dem: NReal; ren_Gen, ren_Dem: NReal;
      CapacidadMaxima_MWh: NReal; fd, TMR: NReal);

    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
    procedure Free; override;
  end;




  TBancoDeBaterias01 = class(TGenerador)
  public
    pa: TFichaBancoDeBaterias01;

    //Variable de Estado
    NDisc: integer; // discretizacion del estad
    Carga_ini: NReal; // valor inicial

    X_Carga: NReal;
    Xs_Carga: NReal;

    cv_Dem, cv_Gen: NReal; // costo variable visto por el despacho

    // derivada del costo futuro respecto a la carga de la bateria
    dCFdC: NReal;

    Disponible: boolean;

    PMax_Gen, PMax_Dem: NReal;



    constructor Create(capa: integer; nombre: string; nacimiento, muerte: TFecha;
      lpdUnidades, lpd: TFichasLPD; nodo: TNodo;
  flg_CalcularGradienteDeInversion: boolean; Carga_ini: NReal; NDisc_: integer;
  TonCO2xMWh: NReal; LowCostMustRun, CleanDevelopmentMechanism: boolean);

    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;



    procedure PrepararMemoria(globs: TGlobs); override;

    procedure RegistrarParametrosDinamicos; override;


    function PotenciaFirme: NReal; override;

    function InfoAd_: string; override;

    class function DescClase: string; override;


    //      procedure Sim_Inicio; override;
    procedure Sim_Cronica_Inicio; override;

    function get_pa_FD( kTipoUnidad: integer ): NReal; override;
    function get_pa_TMR( kTipoUnidad: integer ): NReal; override;

    procedure SorteosDelPaso(sortear: boolean); override;
    procedure PrepararPaso_ps; override;

    procedure opt_nvers(var ivar, ivae, ires: integer); override;

    procedure opt_cargue(s: TSimplex); override;
    procedure opt_fijarRestriccionesDeCaja(s: TSimplex); override;

    procedure opt_leerSolucion(s: TSimplex); override;
    procedure EvolucionarEstado; override;


    function getNombreVar(ivar: integer; var nombre: string): boolean; override;
    function getNombreRes(ires: integer; var nombre: string): boolean; override;

    procedure PosicionarseEnEstrellita; override;

    procedure ActualizarEstadoGlobal( flg_Xs: boolean ); override;

    procedure optx_nvxs(var ixr, ixd, iauxr, iauxd: integer); override;
    procedure optx_RegistrarVariablesDeEstado(adminEstados: TAdminEstados);
      override;

    procedure Free; override;

    procedure PubliVars; override;

    // imprime tabla de valores del agua en USD/Hm3
    procedure opt_PrintResultados_Encab(var fsal: textfile); override;
    procedure opt_PrintResultados(var fsal: textfile); override;

    procedure dump_Variables(var f: TextFile; charIndentacion: char); override;
  end;

procedure AlInicio;
procedure AlFinal;

implementation


constructor TBancoDeBaterias01.Create(capa: integer; nombre: string;
  nacimiento, muerte: TFecha; lpdUnidades, lpd: TFichasLPD; nodo: TNodo;
  flg_CalcularGradienteDeInversion: boolean;
  Carga_ini: NReal; NDisc_: integer; TonCO2xMWh: NReal;
  LowCostMustRun, CleanDevelopmentMechanism: boolean);

begin
  inherited Create(capa, nombre, nacimiento, muerte, lpdUnidades, nodo,
    TonCO2xMWh, LowCostMustRun, CleanDevelopmentMechanism,
      flg_CalcularGradienteDeInversion );
  Self.lpd := lpd;

  self.Carga_ini := Carga_ini;
  self.NDisc := NDisc_;
end;

constructor TBancoDeBaterias01.Create_ReadFromText(f: TArchiTexto);
begin
  lpd := nil;
  inherited Create_ReadFromText(f);
  f.IniciarLecturaRetrasada;
  f.rd('Carga_ini', Carga_Ini);
  f.rd('NDisc', NDisc);
  f.rd('lpd', TCosa(lpd));

  f.EjecutarLectura;
  pa := nil;
  nodo := nil;
  if lpd <> nil then
    lpd.Propietario := self;
end;

procedure TBancoDeBaterias01.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wr('Carga_ini', Carga_ini);
  f.wr('NDisc', NDisc);
  f.wr('lpd', TCosa(lpd));
end;

procedure TBancoDeBaterias01.prepararMemoria(globs: TGlobs);
begin
  inherited prepararMemoria(globs);
end;

procedure TBancoDeBaterias01.registrarParametrosDinamicos;
begin
  inherited registrarParametrosDinamicos;
  lpd.expandirFichas(globs);
  lpd.RegistrarFichasAActualizar(Self, globs.ActualizadorLPD, @pa, nil);
end;

function TBancoDeBaterias01.infoAd_: string;
begin
  Result := '';
end;

function TBancoDeBaterias01.potenciaFirme: NReal;
begin
  Result := 0;
end;

class function TBancoDeBaterias01.DescClase: string;
begin
  Result := rsBancoDeBaterias01;
end;

function TBancoDeBaterias01.get_pa_FD( kTipoUnidad: integer ): NReal;
begin
  result:= pa.fd;
end;

function TBancoDeBaterias01.get_pa_TMR( kTipoUnidad: integer ): NReal;
begin
  result:= pa.tmr;
end;



procedure TBancoDeBaterias01.SorteosDelPaso(sortear: boolean);
begin
  if globs.ObligarDisponibilidad_1_ then
  begin
    Disponible := True;
    PMax_Gen := pa.PMax_Gen;
    PMax_Dem := pa.PMax_Dem;
  end
  else
  begin
    if (sortear) then
    begin
      ActualizarProbabilidadesReparacionYRotura_(pa.fd, pa.tmr);
      Disponible := Sorteos_RepRotUnidades > 0;
      if Disponible then
      begin
        PMax_Gen := pa.PMax_Gen;
        PMax_Dem := pa.PMax_Dem;
      end
      else
      begin
        PMax_Gen := 0;
        PMax_Dem := 0;
      end;
    end
    else
    begin
      Disponible := True;
      PMax_Gen := pa.PMax_Gen * pa.fd;
      PMax_Dem := pa.PMax_Dem * pa.fd;
    end;
  end;
end;

procedure TBancoDeBaterias01.opt_PrintResultados_Encab(var fsal: textfile);
begin
  Write(fsal, #9, FloatToStrF(X_Carga, ffgeneral, 6, 3));
end;

procedure TBancoDeBaterias01.opt_PrintResultados(var fsal: textfile);
begin
  Write(fsal, #9, FloatToStrF(cv_Gen, ffgeneral, 6, 3));
end;

procedure TBancoDeBaterias01.dump_Variables(var f: TextFile; charIndentacion: char);
begin
  inherited dump_Variables(f, charIndentacion);
  writeln(f, charIndentacion, 'Carga[MWh]= ', FloatToStrF(X_Carga, ffFixed, 10, 3));
  writeln(f);
end;

procedure TBancoDeBaterias01.prepararPaso_ps;
begin
  {$IFDEF DEV_CF_ESTRELLA}
  if globs.EstadoDeLaSala = CES_OPTIMIZANDO then
    globs.CF.devxr_estrella_20_(ixr, globs.kPaso_ + 1, cv_Dem, cv_Gen)
  else
  {$ENDIF}
    globs.CF.devxr_continuo_20(ixr, globs.kPaso_ + 1, cv_Dem, cv_Gen);

  cv_Dem := -cv_Dem;
  cv_Gen := -cv_Gen;

end;

procedure TBancoDeBaterias01.opt_nvers(var ivar, ivae, ires: integer);
begin
  self.ivar := ivar;
  self.ires := ires;
  if Disponible then
  begin
    ivar := ivar + 2 * globs.NPostes; // PGen + PDem  por poste
    ires := ires + 2;  // 0 <= EImp <= EImpMax  (segun la fase una de estas)
  end;
end;

procedure TBancoDeBaterias01.opt_cargue(s: TSimplex);
var
  ibaseres: integer;
  iposte: integer;
  jres: integer;
  maux: NReal;
begin

  if not Disponible then
    exit;

{$IFDEF SPXCONLOG}
  spx_NombrarVariables(s);
{$ENDIF}

  ibaseres := nodo.ires;

  // Tenemos que aportar a las restricciones de demanda del nodo al que está
  // conectado el generador
  for iposte := 0 to globs.NPostes - 1 do
  begin
    s.pon_e(ibaseres + iposte, ivar + iposte, 1);
  end;
  for iposte := 0 to globs.NPostes - 1 do
  begin
    s.pon_e(ibaseres + iposte, ivar + globs.NPostes + iposte, -1);
  end;

  // Xs_Carga >= 0
  for iposte := 0 to globs.NPostes - 1 do
    s.pon_e(ires, ivar + iposte, -globs.durpos[iposte] / pa.ren_Gen);

  for iposte := 0 to globs.NPostes - 1 do
    s.pon_e(ires, ivar + globs.NPostes + iposte, -globs.durpos[iposte] * pa.ren_Dem);

  s.pon_e(ires, s.nc, X_Carga);

  // Xs_Carga <= CargaMax.
  for iposte := 0 to globs.NPostes - 1 do
    s.pon_e(ires + 1, ivar + iposte, -s.e(ires, ivar + iposte));
  for iposte := 0 to globs.NPostes - 1 do
    s.pon_e(ires + 1, ivar + globs.NPostes + iposte, -s.e(
      ires, ivar + globs.NPostes + iposte));
  s.pon_e(ires + 1, s.nc, pa.CapacidadMaxima_MWh - s.e(ires, s.nc));

end;

procedure TBancoDeBaterias01.opt_fijarRestriccionesDeCaja(s: TSimplex);
var
  iposte: integer;
begin
  // Le fijamos como cota máxima PMax a la potencia en todos los postes
  if Disponible then
  begin
    for iposte := 0 to globs.NPostes - 1 do
      s.cota_sup_set(ivar + iposte, PMax_Gen);

    for iposte := 0 to globs.NPostes - 1 do
      s.cota_sup_set(ivar + globs.NPostes + iposte, PMax_Dem);
  end;
end;

procedure TBancoDeBaterias01.Sim_Cronica_Inicio;
begin
  inherited Sim_Cronica_Inicio;

  X_Carga := Carga_ini;

end;

procedure TBancoDeBaterias01.opt_leerSolucion(s: TSimplex);
var
  iposte: integer;
begin
  CostoDirectoDelPaso := 0;

  if not disponible then
  begin
    vclear(P);
    Xs_Carga := X_Carga;
    exit;
  end;

  for iposte := 0 to globs.NPostes - 1 do
    P[iposte] := s.xval(ivar + iposte) - s.xval(ivar + globs.npostes + iposte);

  Xs_Carga := s.yval(ires);

end;


procedure TBancoDeBaterias01.EvolucionarEstado;
begin
  X_Carga := Xs_Carga;
end;


function TBancoDeBaterias01.getNombreVar(ivar: integer; var nombre: string): boolean;
begin
  if (ivar >= self.ivar) and (ivar < self.ivar + globs.NPostes) then
  begin
    nombre := self.Nombre + '_P[MW]' + IntToStr(ivar - self.ivar + 1);
    Result := True;
  end
  else
    Result := False;
end;

function TBancoDeBaterias01.getNombreRes(ires: integer; var nombre: string): boolean;
var
  k: integer;
begin
  if (ires = self.ires) then
    nombre := self.nombre + 'Carga >= 0'
  else if (ires = (self.ires + 1)) then
    nombre := self.nombre + 'CapacidadMaxima - Carga >= 0'
  else
    Result := False;
end;

procedure TBancoDeBaterias01.PubliVars;
begin
  inherited PubliVars;
  PublicarVariableNR('cv_Dem', '[USD/MWh]', 6, 1, cv_Dem, True);
  PublicarVariableNR('cv_Gen', '[USD/MWh]', 6, 1, cv_Gen, True);
  PublicarVariableNR('X_Carga', '[MWh]', 6, 1, X_Carga, True);
  PublicarVariableB('Disponible', '-', Disponible, False);
end;

procedure TBancoDeBaterias01.optx_nvxs(var ixr, ixd, iauxr, iauxd: integer);
begin
  self.ixr := ixr;
  ixr := ixr + 1;
end;

procedure TBancoDeBaterias01.PosicionarseEnEstrellita;
begin
  X_Carga := globs.CF.xr[ixr];
end;

procedure TBancoDeBaterias01.ActualizarEstadoGlobal( flg_Xs: boolean );
begin
  globs.CF.xr[ixr] := X_Carga;
end;

procedure TBancoDeBaterias01.optx_RegistrarVariablesDeEstado(
  adminEstados: TAdminEstados);
begin
  adminEstados.Registrar_Continua(
    ixr,
    0.0,
    pa.CapacidadMaxima_MWh,
    ndisc,
    nombre + '_Carga', // nombre de la variable
    'MWh' // unidades
    );
end;

procedure TBancoDeBaterias01.Free;
begin
  if lpd <> nil then
    lpd.Free;
  inherited Free;
end;



{*********************************
*Métodos de TFichaBancoBaterias *
*********************************}

constructor TFichaBancoDeBaterias01.Create(capa: integer; fecha: TFecha;
  periodicidad: TPeriodicidad; PMax_Gen, PMax_Dem: NReal; ren_Gen, ren_Dem: NReal;
  CapacidadMaxima_MWh: NReal; fd, TMR: NReal);
begin
  inherited Create(capa, fecha, periodicidad);
  Self.PMax_Gen := PMax_Gen;
  self.PMax_Dem := PMax_Dem;
  self.ren_Gen := ren_Gen;
  self.ren_Dem := ren_Dem;
  self.CapacidadMaxima_MWh := CapacidadMaxima_MWh;
  self.fd := fd;
  self.TMR := TMR;
end;

procedure TFichaBancoDeBaterias01.WriteToText(f: TArchiTexto);
begin

  inherited WriteToText(f);

  f.wr('PMax_Gen', PMax_Gen);
  f.wr('PMax_Dem', PMax_Dem);
  f.wr('ren_Gen', ren_Gen);
  f.wr('ren_Dem', ren_Dem);
  f.wr('CapacidadMaxima_MWh', CapacidadMaxima_MWh);
  f.wr('fd', fd);
  f.wr('TMR', TMR);
end;

constructor TFichaBancoDeBaterias01.Create_ReadFromText(f: TArchiTexto);
begin
  inherited Create_ReadFromText(f);

  f.IniciarLecturaRetrasada;
  f.rd('PMax_Gen', PMax_Gen);
  f.rd('PMax_Dem', PMax_Dem);
  f.rd('ren_Gen', ren_Gen);
  f.rd('ren_Dem', ren_Dem);
  f.rd('CapacidadMaxima_MWh', CapacidadMaxima_MWh);
  f.rd('fd', fd);
  f.rd('TMR', TMR);
  f.EjecutarLectura;
end;

procedure TFichaBancoDeBaterias01.Free;
begin
  inherited Free;
end;


procedure AlInicio;
begin
  registrarClaseDeCosa(TBancoDeBaterias01.ClassName, TBancoDeBaterias01);
  registrarClaseDeCosa(TFichaBancoDeBaterias01.ClassName, TFichaBancoDeBaterias01);
end;

procedure AlFinal;
begin
end;

end.
