unit uregasificadora;

interface

uses
  uCosa, SysUtils,
  uGlobs, uActores, uFechas, unodos, uCosaConNombre,
  uFichasLPD, usimplex, xMatDefs, uauxiliares, unodocombustible,
  uFuentesAleatorias, uTSumComb, uEstados, uAgendaGNL, Math, uDatosHorariosDetallados;

resourcestring
  rsRegasificadora =
    'Regasificadora';


const
  factor_GNL_a_GN = 610.0;
  C_LARGO_X_SpotYDevios = 6; // pasos_preaviso (6 para test).

  {

**************************************************************
1 cubic meter liquid,  CuM,  600 cubic meter of gas
1 million metric ton/annum,  MMTPA,  4 MMSCMD
1 metric ton LNG,  MT,  1420 cubic meter of gas
1 metric ton LNG,  MT,  52 MMBTU
1000 cubic meter of gas,  MCM,  40 MMBTU
1 british thermal unit,  BTU,  252 calories
**************************************************************
MMSCMD stands for million standard cubic meters per day
MMBTU stands for million British thermal unit
MT stands for metric ton
MMTPA means million metric tons per annum
***************************************************************

  }



type
  TFichaRegasificadora = class(TFichaLPD)


    VGNL_max: NReal; //volumen maximo del tanque
    VGNL_min: NReal; //volumen minimo de GNL que debe quedar en el tanque.

    QGN_BOG_ca: NReal; // BOG [m3/s] constante
    QGN_BOG_cb: NReal; // BOG [1/s] factor aplicable al volumen almacenado
    // QGNL_BOG:= ( BOG_ca + BOG_cb * X_GNL )/ 600

    renRegas: NReal;
    // [p.u.] Rendimiento de regasificación  GN_m3/GN_GNL ; del orden de: 0.95*600
    QGN_Max: NReal; // [m3/s] máximo de Senout de la regasificadora.
    QGN_ConsumosPropios: NReal; // [m3/s] Consumos propios.

    fd: NReal;
    TMR: NReal;
    fuente_idxPrecioCombustible: TFuenteAleatoria;
    borne_idxPrecioCombustible: string;
    kborne_idxPrecioCombustible: integer;

    constructor Create(capa: integer; fecha: TFecha; periodicidad: TPeriodicidad;
      xFuente_idxPrecioCombustible: TFuenteAleatoria;
      xBorne_idxPrecioCombustible: string; fd, TMR: NReal;
      VGNL_min, VGNL_max: NReal;
      QGN_BOG_ca, QGN_BOG_cb, renRegas, QGN_Max, QGN_ConsumosPropios: NReal);

    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
    function InfoAd_: string; override;
  end;


  TRegasificadora = class(TSuministroCombustible)
  public

    // variables
    pa: TFichaRegasificadora;

    Agenda: TAgendaCargosGNL;
    VGNL_inicial: NReal; //volumen inicial del tanque, antes de empezar la simulacion
    X_SpotYDesvios_inicial: TDAOfBoolean;
    Ndiscret: integer;   //Nro de discretizaciones del tanque

    cvVert: NReal;
    //precio al que voy a vender el GNL en caso de que no entre en la regasificadora

    VGNL_Cargo: NReal; // volumen del cargo de GNL [m3]

    precio_Spot: NReal; // precio de los cargos Spot
    precio_Desvio: NREal; // precio de los cargos Desviados
    precio_Agenda: NReal; // precio de los cargos de Agenda



    //volumenes de GNL al inicio y al final del paso
    X_VGNL, Xs_VGNL: NReal; //Xs es final    10 discretizaciones

    //V de vertimiento
    Vvert: NReal;


    // Representación del estado de las decisiones.
    // El casillero CERO es la acción correspondiente a la semana en curso y que
    // fue tomada nPasosDePreAviso hacia atras.
    X_SpotYDesvios: TDAOfBoolean;   // 90 dias = 3 x 4 = 12 semanas 2**12

    // Vector con casillero TRUE si hay una acción (Compra Spot o Desvío)
    Xs_SpotYDesvios: TDAOfBoolean;
    // Vector al final del paso. Es un Shift del X más la decisión del paso.


    // variables auxiliares

    VGNL_minaux: NReal;
    VGNL_maxaux: NReal;

    kCasilleroAgenda: integer; // -1 indica que está fuera de la agenda especificada
    flg_CargoAgendado: boolean; // Indica si en la Agenda hay un cargo para esta semana.
    flg_Desvio: boolean; // en esta semana se produce un desvió de un cargo de Agenda
    flg_Spot: boolean; // en esta semana

    V_Descargado, V_Desviado: NReal; // Descargas de GNL en el paso de tiempo
    CostoCargo: NReal; // Costo directo por Descarga y/o Desvio imputable al paso.
    cv_GNL: NReal; // Valor futuro del GNL Almacenado
    dCF_Accion: NReal; // Valor del futuro por tomar una acción NPasosPrevisoAntes.

    QGN_BOG, VGNL_BOG: NReal;
    // Caudal de Boil Off Gas (BOG) GN y Volumen equivalente de GNL en el paso

    ArchiDatos: TArchiRef;
    datos: TDatosHorariosDetallados;


    constructor Create(capa: integer; nombre: string; nacimiento, muerte: TFecha;
      lpdUnidades: TFichasLPD; lpd: TFichasLPD; NodoSumComb: TNodoCombustible;
      Agenda: TAgendaCargosGNL; VGNL_inicial: NReal; Ndiscret: integer;
      cvVert, VGNL_Cargo, precio_Spot, precio_Desvio, precio_Agenda: NReal);

    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
    procedure PubliVars; override;

    class function DescClase: string; override;

    procedure Sim_Cronica_Inicio; override;

    procedure optx_RegistrarVariablesDeEstado(adminEstados: TAdminEstados); override;
    procedure EvolucionarEstado; override;

    procedure PrepararMemoria(globs: TGlobs); override;
    procedure RegistrarParametrosDinamicos; override;

    procedure PrepararPaso_ps; override;

    procedure optx_nvxs(var ixr, ixd, iauxr, iauxd: integer); override;

    procedure ActualizarEstadoGlobal( flg_Xs: boolean ); override;
    procedure PosicionarseEnEstrellita; override;

    procedure opt_nvers(var ivar, ivae, ires: integer); override;
    procedure opt_cargue(s: TSimplex); override;
    procedure opt_fijarRestriccionesDeCaja(s: TSimplex); override;
    procedure opt_leerSolucion(s: TSimplex); override;

    procedure CambioFichaPD; override;


    function getNombreVar(ivar: integer; var nombre: string): boolean; override;
    function getNombreRes(ires: integer; var nombre: string): boolean; override;

    procedure Free; override;

    function X_SpotYDesvios_Inicial_AsBoolStr(nbinsPorLinea: integer): string;
    procedure Set_X_SpotYDesvios_inicial_FromBoolStr(XBoolStr: string);

  end;


function ceros_str(n: integer): string;

procedure TRegasificadora_CambioFichaPD(actor: TCosa);

procedure AlInicio;
procedure ALFinal;

implementation

function ceros_str(n: integer): string;
var
  k: integer;
  s: string;
begin
  setlength(s, n);
  for k := 1 to n do
    s[k] := '0';
  Result := s;
end;


procedure TRegasificadora_CambioFichaPD(actor: TCosa);
begin
  (actor as TRegasificadora).CambioFichaPD;
end;

// Métodos de TFichaRegasificadora

constructor TFichaRegasificadora.Create(capa: integer; fecha: TFecha;
  periodicidad: TPeriodicidad; xFuente_idxPrecioCombustible: TFuenteAleatoria;
  xBorne_idxPrecioCombustible: string; fd, TMR: NReal; VGNL_min, VGNL_max: NReal;
  QGN_BOG_ca, QGN_BOG_cb, renRegas, QGN_Max, QGN_ConsumosPropios: NReal);

begin
  inherited Create(capa, fecha, periodicidad);
  self.fd := fd;
  self.TMR := TMR;
  self.fuente_idxPrecioCombustible := xFuente_idxPrecioCombustible;
  self.borne_idxPrecioCombustible := xBorne_idxPrecioCombustible;
  self.VGNL_min := VGNL_min;
  self.VGNL_max := VGNL_max;
  self.QGN_BOG_ca := QGN_BOG_CA;
  self.QGN_BOG_cb := QGN_BOG_cb;
  self.renRegas := renRegas;
  self.QGN_Max := QGN_Max;
  self.QGN_ConsumosPropios := QGN_ConsumosPropios;
end;

constructor TFichaRegasificadora.Create_ReadFromText(f: TArchiTexto);
begin
  inherited Create_ReadFromText(f);
  f.IniciarLecturaRetrasada;
  f.rd('fd', fd);
  f.rd('TMR', TMR);
  f.rdReferencia('fuente_idxPrecioCombustible',
    TCosa(fuente_idxPrecioCombustible), self);
  f.rd('borne_idxPrecioCombustible', borne_idxPrecioCombustible);
  f.rd('VGNL_min', VGNL_min);
  f.rd('VGNL_max', VGNL_max);
  f.rd('QGN_BOG_ca', QGN_BOG_ca);
  f.rd('QGN_BOG_cb', QGN_BOG_cb);
  f.rd('renRegas', renRegas);
  f.rd('QGN_Max', QGN_Max);
  f.rd('QGN_ConsumosPropios', QGN_ConsumosPropios);
  f.EjecutarLectura;
end;

procedure TFichaRegasificadora.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wr('fd', fd);
  f.wr('TMR', TMR);
  f.wrReferencia('fuente_idxPrecioCombustible', fuente_idxPrecioCombustible);
  f.wr('borne_idxPrecioCombustible', borne_idxPrecioCombustible);
  f.wr('VGNL_min', VGNL_min);
  f.wr('VGNL_max', VGNL_max);
  f.wr('QGN_BOG_ca', QGN_BOG_ca);
  f.wr('QGN_BOG_cb', QGN_BOG_cb);
  f.wr('renRegas', renRegas);
  f.wr('QGN_Max', QGN_Max);
  f.wr('QGN_ConsumosPropios', QGN_ConsumosPropios);
end;

function TFichaRegasificadora.InfoAd_: string;
begin
  Result := 'QGN_Max: ' + FloatToStr(QGN_Max) + ', VGNL_max:' + FloatToStr(VGNL_max);
end;



//--------------------------
// Métodos de TRegasificadora
//==========================

constructor TRegasificadora.Create(capa: integer; nombre: string;
  nacimiento, muerte: TFecha; lpdUnidades: TFichasLPD; lpd: TFichasLPD;
  NodoSumComb: TNodoCombustible; Agenda: TAgendaCargosGNL; VGNL_inicial: NReal;
  //volumen inicial del tanque, antes de empezar la simulacion
  Ndiscret: integer;   //Nro de discretizaciones del tanque
  cvVert, VGNL_Cargo, precio_Spot, precio_Desvio, precio_Agenda: NReal);
begin
  inherited Create(capa, nombre, nacimiento, muerte, lpdUnidades, NodoSumComb);
  self.lpd := lpd;
  self.Agenda := Agenda;
  self.VGNL_inicial := VGNL_inicial;
  self.Ndiscret := Ndiscret;
  self.cvVert := cvVert;
  self.VGNL_Cargo := VGNL_Cargo;
  self.precio_Spot := precio_Spot;
  self.precio_Desvio := precio_Desvio;
  self.precio_Agenda := precio_Agenda;
  setlength(X_SpotYDesvios_inicial, C_LARGO_X_SpotYDevios);
end;

constructor TRegasificadora.Create_ReadFromText(f: TArchiTexto);
begin
  inherited Create_ReadFromText(f);
  f.IniciarLecturaRetrasada;
  f.rd('Agenda', TCosa(agenda));
  f.rd('VGNL_inicial', VGNL_inicial);
  f.rd('cvVert', cvVert);
  f.rd('Ndiscret', Ndiscret);

  if f.version >= 117 then
  begin
    f.rd('VGNL_Cargo', VGNL_Cargo);
    f.rd('precio_Spot', precio_Spot);
    f.rd('precio_Desvio', precio_Desvio);
    f.rd('precio_Agenda', precio_Agenda);
  end
  else
  begin
    VGNL_Cargo := 145000;
    precio_Spot := 18;
    precio_Desvio := 7;
    precio_Agenda := 13;
  end;

  f.rd('X_SpotYDesvios_inicial', X_SpotYDesvios_inicial);
  f.rd('lpd', TCosa(lpd));
  f.EjecutarLectura;
  lpd.Propietario := self;
end;

procedure TRegasificadora.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wr('Agenda', agenda);
  f.wr('VGNL_inicial', VGNL_inicial);
  f.wr('cvVert', cvVert);
  f.wr('Ndiscret', Ndiscret);
  f.wr('VGNL_Cargo', VGNL_Cargo);
  f.wr('precio_Spot', precio_Spot);
  f.wr('precio_Desvio', precio_Desvio);
  f.wr('precio_Agenda', precio_Agenda);
  f.wr('X_SpotYDesvios_inicial', X_SpotYDesvios_inicial);
  f.wr('lpd', lpd);
end;

procedure TRegasificadora.PubliVars;
begin
  inherited PubliVars;
end;



procedure TRegasificadora.PrepararMemoria(globs: TGlobs);
begin
  inherited PrepararMemoria(globs);
  setlength(X_SpotYDesvios, C_LARGO_X_SpotYDevios);
  setlength(Xs_SpotYDesvios, C_LARGO_X_SpotYDevios);
end;


class function TRegasificadora.DescClase: string;
begin
  Result := rsRegasificadora;
end;

procedure TRegasificadora.RegistrarParametrosDinamicos;
var
  i: integer;
  ficha: TFichaRegasificadora;

begin
  inherited registrarParametrosDinamicos;
  lpd.expandirFichas(globs);
  lpd.RegistrarFichasAActualizar(Self, globs.ActualizadorLPD, @pA,
    nil, TRegasificadora_CambioFichaPD);

  for i := 0 to lpd.Count - 1 do
  begin
    ficha := lpd[0] as TFichaRegasificadora;
    if ficha.fuente_idxPrecioCombustible <> nil then
      ficha.kborne_idxPrecioCombustible :=
        ficha.fuente_idxPrecioCombustible.IdBorne(ficha.borne_idxPrecioCombustible);
  end;
  pa := lpd[0] as TFichaRegasificadora;
  CambioFichaPD;
end;



procedure TRegasificadora.optx_RegistrarVariablesDeEstado(adminEstados: TAdminEstados);
var
  k: integer;
begin
  adminEstados.Registrar_Continua(
    ixr,
    pa.VGNL_min,
    pa.VGNL_max,
    self.NDiscret,
    nombre + '_VolGNL', // nombre de la variable
    'm3' // unidades
    );

  for k := 0 to high(Xs_SpotYDesvios) do
    adminEstados.Registrar_Discreta(ixd + k, 2, nombre + '_SpotYDesvio', 'bool');
end;


procedure TRegasificadora.Sim_Cronica_Inicio;
begin
  X_VGNL := self.VGNL_inicial;
  Xs_VGNL := X_VGNL;
  X_SpotYDesvios := copy(X_SpotYDesvios_inicial);
end;

procedure TRegasificadora.Set_X_SpotYDesvios_inicial_FromBoolStr(XBoolStr: string);
var
  k, n: integer;
  cnt: integer;
begin
  cnt := 0;
  for k := 0 to high(X_SpotYDesvios_Inicial) do
    X_SpotYDesvios_Inicial[k] := False;

  n := length(X_SpotYDesvios_Inicial);
  k := 1;
  while (cnt < n) and (k <= length(XBoolStr)) do
  begin
    if (XBoolStr[k] = '0') then
    begin
      X_SpotYDesvios_Inicial[cnt] := False;
      Inc(cnt);
    end
    else if (XBoolStr[k] = '1') then
    begin
      X_SpotYDesvios_Inicial[cnt] := True;
      Inc(cnt);
    end;
    Inc(k);
  end;
end;


procedure TRegasificadora.opt_nvers(var ivar, ivae, ires: integer);
begin
  self.ivar := ivar;
  Inc(ivar, globs.npostes + 2); //una ivar para cada Q por poste,
  // una para el estado Xs_VGNL y una para el Vvert.

  //la primer restricción es la ecuación de evolución del estado: f(X,u,r) - Xs_VGNL = 0
  // la segunda restricción impone la evacuacion del BOG
  self.ires := ires;
  Inc(ires, 2);
end;


procedure TRegasificadora.PrepararPaso_ps;
var
  dCFdX_inc, dCFdX_dec: NReal;
  rescod: integer;
  xrpos: NReal;
  dCFdX: NReal;
  k: integer;

begin
  // Ahora derivamos la función de CF(X, k+1) respecto al volumen almacenado
  {$IFDEF DEV_CF_ESTRELLA}
  if globs.EstadoDeLaSala = CES_OPTIMIZANDO then
    globs.CF.devxr_estrella_(ixr, globs.kPaso_ + 1, dCFdX_inc, dCFdX_dec, rescod)
  else
  {$ENDIF}
    globs.CF.devxr_continuo(ixr, globs.kPaso_ + 1, dCFdX_inc,
      dCFdX_dec, rescod, xrpos);

  case rescod of
    -1: dCFdX := dCFdX_inc;
    0: dCFdX := (dCFdX_dec + dCFdX_inc) / 2.0;
    1: dCFdX := dCFdX_dec;
  end;

  // El vlor del GNL almacenado es menos la derivada del costo futuro respecto
  // al volumen almacenado y por el factor de actualización del paso.
  cv_GNL := -dCFdX * globs.fActPaso;

  // Consultamos si en La Agenda (original) hay un cargo agendado para
  // este paso.
  kCasilleroAgenda := Agenda.kCasilleroAgenda(globs.kPaso_);
  flg_CargoAgendado := Agenda.HayCargoAgendado(kCasilleroAgenda);



  // Ahora según la tabla de deciciones anteriores que afectan este Casillero
  // determinamos si arriva o no un cargo si el mismo es de Agenda o Spot y si
  // hay o no desvió de un cargo.
  if flg_CargoAgendado then
  begin
    flg_Spot := False;
    // Hay un cargo en la Agenda por lo cual, si en la tabla de decisiones hay una
    // acción para este casillero, se trata de DESVIAR el cargo
    if (X_SpotYDesvios[0]) then
    begin  // El cargo fue desviado
      flg_Desvio := True;
      CostoCargo := Agenda.CostoDesvio;
      V_Descargado := 0;
      V_Desviado := Agenda.VGNL_m3;
    end
    else
    begin // El cargo es descargado en la regasificadora
      flg_Desvio := False;
      CostoCargo := Agenda.CostoAgenda;
      V_Descargado := Agenda.VGNL_m3;
      V_Desviado := 0;
    end;
  end
  else
  begin
    flg_Desvio := False;
    // NO hay un cargo Agendado, por lo cual, si en la tabla de decisiones hay una
    // acción para esta semana, se trata de una compra Spot. Sino no hay descarga.
    if (X_SpotYDesvios[0]) then
    begin   // Se descarga un cargo Spot
      flg_Spot := True;
      CostoCargo := Agenda.CostoSpot;
      V_Descargado := Agenda.VGNL_m3;
      V_Desviado := 0;
    end
    else
    begin   // Semana sin descargas ni desvios
      flg_Spot := False;
      CostoCargo := 0;
      V_Descargado := 0;
      V_Desviado := 0;
    end;
  end;

  if pa.fuente_idxPrecioCombustible <> nil then
    CostoCargo := CostoCargo *
      pa.fuente_idxPrecioCombustible.Bornera[pa.kborne_idxPrecioCombustible];


  // Cálculo del BOIL OFF
  QGN_BOG := pa.QGN_BOG_ca + pa.QGN_BOG_cb * X_VGNL;
  VGNL_BOG := (QGN_BOG * globs.SegundosDelPaso) / factor_GNL_a_GN;

  //Si lo que queda de GNL es menor que el V_min no quiero sacar mas, para eso aumento el precio
  if (X_VGNL <= pa.VGNL_min) then
    cv_GNL := 10 * cv_GNL;      { TODO : -- revisar criterio ... }

  // Caudal de Boil Off Gas (BOG) y Volumen en el paso

  if X_VGNL + V_Descargado - VGNL_BOG < 0 then
    VGNL_BOG := X_VGNL + V_Descargado;



  // Cálculo de la afectación del futuro por decidir cambiar lo agendado para dentro de NPreavisoSemanas
  {$IFDEF DEV_CF_ESTRELLA}
  if globs.EstadoDeLaSala = CES_OPTIMIZANDO then
    dCF_accion := globs.CF.deltaCosto_vxd_estrella_(
      ixd + High(X_SpotYDesvios), globs.kPaso_ + 1, 1) * globs.fActPaso
  else
  {$ENDIF}
    dCF_accion := globs.CF.deltaCosto_vxd_continuo(ixd + High(
      X_SpotYDesvios), globs.kPaso_ + 1, 1) * globs.fActPaso;

  for k := 1 to high(X_SpotYDesvios) do
    Xs_SpotYDesvios[k - 1] := X_SpotYDesvios[k];
  Xs_SpotYDesvios[ high( Xs_SpotYDesvios )]:= false;

  {Toma la decisión suponiendo que los costos directos de la misma son
  pagados en la semana futura donde tiene consecuencias. (si se agrega penalidad
  por desvío esa tendría que ir aquí.}
  if dCF_Accion < 0 then
    Xs_SpotYDesvios[high(Xs_SpotYDesvios)] := True
  else
    Xs_SpotYDesvios[high(Xs_SpotYDesvios)] := False;

end;

procedure TRegasificadora.opt_cargue(s: TSimplex);
var
  iresnodo: integer;
  kposte: integer;
  VAporte: NReal;
  dCFdV_Inc, dCFdV_Dec: NReal;
  resCode: integer;

begin
  iresnodo := self.NodoComb.ires;
  for kposte := 0 to globs.npostes - 1 do
  begin
    // En cada restricción del nodo_combustible aportamos
    // el caudal despachado.
    s.pon_e(iresnodo + kposte, ivar + kposte, 1);

    // Consideramos las extracciones asociadas a los caudales despachados
    // en la restricción dinámica f(X, u, r ) - Xs = 0
    // ( Aportes - SUM(Qi*durposi/ 600)  - Xs_VGNL + (X_VGNL - Vvert) = 0)
    s.pon_e(ires, ivar + kposte, -globs.DurPos[kposte] / factor_GNL_a_GN / pa.renRegas );
  end;

  //cargar termino de Xs_VGNL
  s.pon_e(ires, ivar+ globs.npostes + 0, -1);

  //cargar termino de Vvert
  s.pon_e(ires, ivar+ globs.npostes + 1, -1);

  //cargar termino constante
  s.pon_e(ires, s.nc, self.X_VGNL + V_Descargado);


  // Cargado de la restricción SUM( Qi*durposi/ 600 ) + VVert - VGNL_BOG >=0
  for kposte := 0 to globs.npostes - 1 do
  begin
    s.pon_e(ires + 1, ivar + kposte, -s.e(ires, ivar + kposte));
  end;
  s.pon_e(ires + 1, ivar + globs.npostes + 1, 1);
  s.pon_e(ires + 1, s.nc, -VGNL_BOG);
  { TODO : ojo revisar ,, que pasa si hay vertimiento porque no entra el cargo en la regasificador cómo considero el BOG en ese caso Capaz que no importa pues se pierde igual }

  // cargar funcion objetivo -costo (derivada de CF)
  s.pon_e(s.nf, ivar + globs.npostes + 0, -cv_GNL);

  // cargar funcion ingreso por ventas de vertimiento
  s.pon_e(s.nf, ivar + globs.npostes + 1, cvVert);

  // cargar funcion objetivo -costoCargo (Agenda, Spot o Agenda-Desvio )
  s.acum_e(s.nf, s.nc, -CostoCargo);

end;




procedure TRegasificadora.opt_fijarRestriccionesDeCaja(s: TSimplex);
var
  kposte: integer;
begin
  s.FijarRestriccionIgualdad(ires);

  for kposte := 0 to globs.npostes - 1 do
    s.cota_sup_set(ivar + kposte, pa.QGN_Max);

  s.cota_sup_set(ivar + globs.npostes + 1, pa.VGNL_max);//  Xs_VGNL< VGNL_max
  s.cota_inf_set(ivar + globs.npostes + 1, 0);//  Xs_VGNL> 0
end;



procedure TRegasificadora.EvolucionarEstado;

begin
  X_VGNL := Xs_VGNL;
  vswap(X_SpotYDesvios, Xs_SpotYDesvios);
end;

procedure TRegasificadora.optx_nvxs(var ixr, ixd, iauxr, iauxd: integer);
begin
  self.ixr := ixr;
  ixr := ixr + 1;
  self.ixd := ixd;
  Inc(ixd, length(X_SpotYDesvios));
end;

procedure TRegasificadora.PosicionarseEnEstrellita;
var
  iRetardo: integer;

begin
  X_VGNL := globs.CF.xr[ixr];
  for iRetardo := 0 to high(X_SpotYDesvios) do
    X_SpotYDesvios[iRetardo] := (globs.CF.xd[ixd + iRetardo]) <> 0;
end;



procedure TRegasificadora.ActualizarEstadoGlobal( flg_Xs: boolean );
var
  iRetardo: integer;
begin
  globs.CF.xr[ixr] := Xs_VGNL;

  if flg_Xs then
  begin
    for iRetardo := 0 to high(X_SpotYDesvios)-1 do
    begin
      if X_SpotYDesvios[iRetardo+1] then
        globs.CF.xd[ixd + iRetardo] := 1
      else
        globs.CF.xd[ixd + iRetardo] := 0;
    end;
    globs.CF.xd[ixd + high( X_SpotYDesvios )] := 0;
  end
  else
    for iRetardo := 0 to high(X_SpotYDesvios) do
    begin
      if X_SpotYDesvios[iRetardo] then
        globs.CF.xd[ixd + iRetardo] := 1
      else
        globs.CF.xd[ixd + iRetardo] := 0;
    end
end;


procedure TRegasificadora.opt_leerSolucion(s: TSimplex);
var
  kposte: integer;
  Consumo: NREal;
begin
  Consumo := 0;
  for kposte := 0 to globs.npostes - 1 do
  begin
    Q[kposte] := s.xval(ivar + kposte);
    Consumo := Consumo + Q[kposte] * globs.DurPos[kposte];
  end;
  self.Xs_VGNL := s.xval(ivar + globs.NPostes + 1);
  self.Vvert := s.xval(ivar + globs.NPostes + 2);
  CostoDirectoDelPaso := costoCargo;
end;

function Tregasificadora.X_SpotYDesvios_Inicial_AsBoolStr(
  nbinsPorLinea: integer): string;
var
  col, k: integer;
  res: string;
begin
  res := '';
  col := 0;
  for k := 0 to high(X_SpotYDesvios_inicial) do
  begin
    if X_SpotYDesvios_inicial[k] then
      res := res + '1'
    else
      res := res + '0';
    Inc(col);
    if col = nbinsPorLinea then
    begin
      res := res + #13#10;
      col := 0;
    end;
  end;
  Result := res;
end;




procedure TRegasificadora.CambioFichaPD;
begin
(*
   precio_USD_por_Q1h:= NodoComb.combustible.precio_USD_por_MWh_a_PCI(
    pa.precio_por_unidad, pa.unidades_del_precio, pa.flg_precio_a_PCS) * NodoComb.combustible.MWh_por_Q1h;
*)
end;


function TRegasificadora.getNombreVar(ivar: integer; var nombre: string): boolean;
var
  jvar: integer;
begin

  jvar := ivar - self.ivar;
  if (jvar >= 0) and (jvar < (globs.NPostes + 2)) then
  begin
    if jvar < globs.NPostes then
      nombre := self.Nombre + '_Q[m3/s]' + IntToStr(jvar + 1)
    else
    begin
      jvar := jvar - globs.NPostes;
      case jvar of
        0: nombre := self.Nombre + '_Xs_VGNL[m3]';
        1: nombre := self.Nombre + '_VVertGNL[m3]';
      end;
    end;
    Result := True;
  end
  else
    Result := False;
end;


function TRegasificadora.getNombreRes(ires: integer; var nombre: string): boolean;
var
  jRes: integer;
begin
  jRes := ires - self.ires;
  if (jres >= 0) and (jres < 2) then
  begin
    case jRes of
      0: nombre := self.nombre + '_DYN';
      1: nombre := self.nombre + '_BOG';
    end;
    result:= true;
  end
  else
    Result := false;
end;


procedure TRegasificadora.Free;
begin
  setlength(X_SpotYDesvios, 0);
  setlength(Xs_SpotYDesvios, 0);
  inherited Free;
end;

procedure AlInicio;
begin
  registrarClaseDeCosa(TRegasificadora.ClassName,
    TRegasificadora);
  registrarClaseDeCosa(TFichaRegasificadora.ClassName,
    TFichaRegasificadora);
end;

procedure AlFinal;
begin
end;

end.
