{xDEFINE PERTURBADO}
{xDEFINE RESTO_MINCOSTO}
unit usalasdejuegoParaEditor;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface
uses
  classes, sysutils,
  usalasdejuego,
  uActores,
  uCosaConNombre,
  xmatdefs,
  ucosa,
  uauxiliares,
  ufechas,
  uDemandas, uGeneradores,
  uArcos,
  uArcoConSalidaProgramable,
  uNodos,
  uUsoGestionable_postizado,
  uComercioInternacional,
  uFuentesAleatorias,
  unodocombustible,
  uarcocombustible,
  ugsimple_bicombustible,
  uRegasificadora,
  uTSumComb,
  uCombustible,
  uDemandaCombustibleAnioBaseEIndices;

resourcestring
  mesElActor = 'El actor ';
  exNoPerteneceGrupoActores = ' no pertenece a ningun grupo de actores';
  exLaCosaConNombre = 'uSalasDeJuegoParaEditor.addCosaConNombre: La cosa ';
  exNoPerteneceANingunaLista = ' no pertenece a ninguna lista conocida.';


function getListaDeActor( sala: TSalaDeJuego; actor: TCosaConNombre): TListaDeCosasConNombre;

function addCosaConNombre(sala: TSalaDeJuego; actor : TCosaConNombre) : boolean;
//Saca el actor de la sala y libera su memoria
procedure eliminarActor(sala: TSalaDeJuego; var actor : TActor);
//Saca el actor de la sala sin liberar su memoria
procedure quitarActor(sala: TSalaDeJuego; actor : TActor);

function nombreRepetido(sala: TSalaDeJuego; cosaConNombreEditando: TCosaConNombre; nombre: String): boolean;
//tipoActor puede ser 'generador', 'demanda', 'arco' o 'nodo', sino se lanza una excepcion
//retorna true si la lista de actores correspondiente contiene un actor con el mismo nombre
//que no sea actorEditando. Si actorEditando es nil retorna true si la lista
//de actores correspondiente conitene un actor con el mismo nombre
function buscarCosaConNombre(sala: TSalaDeJuego; const nombre: String): TCosaConNombre;
//Resuelve todas las referencias A actores o cosas que se encuentren en la sala.
//No confundir con resuelve las referencias DE actores en la sala.
//si eliminarNoEncontradas es false Retorna la cantidad de referencias que no se resolvieron
//si eliminarNoEncontradas es true, aquellas referencias que no se puedan resolver
//son ELIMINADAS de la lista de referencias, asignadas a NIL, y la función retorna 0
function buscarCosaConNombrePorReferencia(sala: TSalaDeJuego; const referencia: String): TCosaConNombre;
function resolverReferenciasContraSala(sala: TSalaDeJuego ; eliminarNoEncontradas : boolean) : Integer;
//Igual que resolverReferenciasContraSala pero solo resuelve las referencias de cosa
//function resolverReferenciasDeCosaContraSala(cosa : TCosa ; sala : TSalaDeJuego; eliminarNoEncontradas : boolean) : Integer;

function demandaFirme(sala: TSalaDeJuego ): NReal;
function potenciaFirme( sala: TSalaDeJuego ): NReal;
function Clonar_Y_ResolverReferencias(sala: TSalaDeJuego; cosaAClonar : TCosa) : TCosa;

function existeReferenciaALaCosaConNombre(cosaConNombre: TCosaConNombre; sala: TSalaDeJuego) : boolean;

procedure AlInicio;
procedure AlFinal;

implementation

function getListaDeActor(sala: TSalaDeJuego; actor : TCosaConNombre) : TListaDeCosasConNombre;
var
  lista: TListaDeCosasConNombre;
begin
if actor is TGenerador then	lista := sala.Gens_
else if actor is TDemanda then lista := sala.Dems_
else if actor is TNodo then lista := sala.Nods_
else if actor is TArco then lista := sala.Arcs_
else if actor is TArcoConSalidaProgramable then lista := sala.Arcs_
else if (actor is TComercioInternacional) then lista := sala.ComercioInternacional_
else if actor is TUsoGestionable_postizado then lista:= sala.UsosGestionables
else if (actor is TNodoCombustible)
     or (actor is TArcoCombustible)
     or (actor is TGSimple_BiCombustible )
     or (actor is TSuministroCombustible )
     or (actor is TSuministroSimpleCombustible )
     or (actor is TRegasificadora )
     or (actor is TDemandaCombustibleAnioBaseEIndices )
     then lista:= sala.Sums
else
  raise Exception.Create(exLaCosaConNombre + actor.ClaseNombre + exNoPerteneceANingunaLista);
  result:= Lista;
end;


function addCosaConNombre(sala: TSalaDeJuego; actor : TCosaConNombre) : boolean;
var
  aux : Integer;
  lista : TListaDeCosasConNombre;
begin
  if actor is TFuenteAleatoria then
  begin // si es una fuente la agregamos a la lista de fuentes
    lista:= sala.listaFuentes;
    if not lista.find(actor.nombre, aux) then
    begin
      lista.Add(actor);
      result := true;
    end
    else
      result := false;
  end
  else if actor is TCombustible then
  begin
    lista:= sala.listaCombustibles;
    if not lista.find(actor.nombre, aux) then
    begin
      lista.Add(actor);
      result := true;
    end
    else
      result := false;
  end
  else
  begin

    lista:= getListaDeActor( sala, Actor );
    if not lista.find(actor.nombre, aux) then
    begin
      lista.Add(actor);
      sala.listaActores.Add(actor);
      result := true;
    end
    else
      result := false
  end;
end;

procedure eliminarActor(sala: TSalaDeJuego; var actor : TActor);
var
  lista : TListaDeCosas;
begin
  lista:= getListaDeActor(sala, actor);
  lista.Remove(actor);
  sala.listaActores.Remove(actor);
  actor.free;
end;

procedure quitarActor( sala: TSalaDeJuego; actor : TActor);
var
	lista : TListaDeCosas;
begin
  lista:= getListaDeActor(sala, actor);
	lista.Remove(actor);
	sala.listaActores.Remove(actor);
end;

function nombreRepetido(sala: TSalaDeJuego; cosaConNombreEditando: TCosaConNombre; nombre: String): boolean;
var
  i, posCosa: Integer;
begin
  posCosa:= -1;
  for i:= 0 to sala.listaActores.Count -1 do
    if (sala.listaActores[i] <> cosaConNombreEditando) and
       (TActor(sala.listaActores[i]).nombre = nombre) then
    begin
      posCosa:= i;
      break;
    end;
  if posCosa = -1 then
    for i:= 0 to sala.listaFuentes.Count -1 do
      if (sala.listaFuentes[i] <> cosaConNombreEditando) and
         (TFuenteAleatoria(sala.listaFuentes[i]).nombre = nombre) then
      begin
        posCosa:= i;
        break;
      end;
  //TODO faltan los monitores

    if posCosa = -1 then
    for i:= 0 to sala.listaCombustibles.Count -1 do
      if (sala.listaCombustibles[i] <> cosaConNombreEditando) and
         (TCombustible(sala.listaCombustibles[i]).nombre = nombre) then
      begin
        posCosa:= i;
        break;
      end;

  result:= posCosa <> -1;
end;

function buscarCosaConNombre(sala: TSalaDeJuego; const nombre: String) : TCosaConNombre;
var
	cosa : TCosaConNombre;
begin
	if sala.globs.nombre = nombre then
		cosa := sala.globs
	else
  begin
    cosa := sala.listaActores.find(nombre);
    if cosa = NIL then
      cosa := sala.listaFuentes.find(nombre);
  end;
	result := cosa;
end;

function buscarCosaConNombrePorReferencia(sala: TSalaDeJuego; const referencia: String): TCosaConNombre;
var
	cosa : TCosaConNombre;
  clase, nombre: String;
begin
  parsearReferencia(referencia, clase, nombre);

	if (sala.globs.nombre = nombre) and (sala.globs.ClassName = clase) then
		cosa := sala.globs
	else
  begin
    cosa := sala.listaActores.find(nombre);
    if cosa = NIL then
      cosa := sala.listaFuentes.find(nombre);
  end;
	result := cosa;
end;

function resolverReferenciasContraSala(sala : TSalaDeJuego ; eliminarNoEncontradas : boolean) : Integer;
var
  lstGlobs: TListaDeCosasConNombre;
  res : Integer;
begin
  if uCosaConNombre.referenciasSinResolver > 0 then
  begin
    lstGlobs:= TListaDeCosasConNombre.Create(0,'Aux');
    lstGlobs.Add(sala.globs);
    uCosaConNombre.resolver_referencias(lstGlobs);
    uCosaConNombre.resolver_referencias(sala.listaActores);
    uCosaConNombre.resolver_referencias(sala.listaCombustibles);
    res:= uCosaConNombre.resolver_referencias(sala.listaFuentes);
    lstGlobs.FreeSinElemenentos;

    if eliminarNoEncontradas then
      uCosaConNombre.LimpiarReferencias;

    result:= res;
  end
  else
    Result:= 0;
end;

{function resolverReferenciasDeCosaContraSala(cosa : TCosa ; sala : TSalaDeJuego; eliminarNoEncontradas : boolean) : Integer;
var
  cosasEnLaSala : TListaDeCosasConNombre;
  i, n, refsSinResolver : Integer;
begin
  if uCosaConNombre.referenciasSinResolver > 0 then
  begin
    cosasEnLaSala:= TListaDeCosasConNombre.Create('Aux');
    n:= sala.listaActores.Count +
        sala.listaFuentes.Count +
        sala.Funcs.Count;
    cosasEnLaSala.Capacity:= n;
    for i:= 0 to sala.listaActores.Count - 1 do
      cosasEnLaSala.Add(TActor(sala.listaActores[i]));
    for i:= 0 to sala.listaFuentes.Count - 1 do
      cosasEnLaSala.Add(TFuenteAleatoria(sala.listaFuentes[i]));
    for i := 0 to sala.funcs.Count - 1 do
      cosasEnLaSala.Add(TFuncion(sala.Funcs[i]));
//    if cosa is TCosaConNombre then
//      cosasEnLaSala.Add(TCosaConNombre(cosa));
    refsSinResolver:= uCosaConNombre.resolver_referenciasDeCosa(cosa, cosasEnLaSala);
    cosasEnLaSala.FreeSinElemenentos;
    if eliminarNoEncontradas then
      uCosaConNombre.eliminar_referencias_del(cosa);

    result:= refsSinResolver;
  end
  else
    Result:= 0;
end;}

function demandaFirme (sala: TSalaDeJuego ): NReal;
var
	i : Integer;
	resultado : NReal;
begin
	resultado := 0;
	if sala.Dems_.Count > 0 then
		begin
		for i := 0 to sala.Dems_.Count -1 do
			resultado := resultado + (sala.dems_[i] as TActor).potenciaFirme;
		end;
	result := resultado
end;

function potenciaFirme(sala: TSalaDeJuego ): NReal;
var
	i : Integer;
	resultado : NReal;
begin
	resultado := 0;
	for i := 0 to sala.gens_.Count -1 do
		resultado := resultado + (sala.gens_[i] as TGenerador ).potenciaFirme;
	Result := resultado
end;

function Clonar_Y_ResolverReferencias(sala: TSalaDeJuego; cosaAClonar : TCosa) : TCosa;
var
  aux : TCosa;
begin
  assert(uCosaConNombre.referenciasSinResolver = 0);
  aux := cosaAClonar.Create_Clone;
  resolverReferenciasContraSala(sala, false);
  assert(uCosaConNombre.referenciasSinResolver = 0);
  result := aux;
end;

function existeReferenciaALaCosaConNombre(cosaConNombre: TCosaConNombre; sala: TSalaDeJuego) : boolean;
var
  salaAux: TSalaDeJuego;
  res: boolean;
begin
  assert(uCosaConNombre.referenciasSinResolver = 0, 'usalasdejuegoParaEditor.existeReferenciaALaCosaConNombre: quedan referencias sin resolver en la sala');
  salaAux:= TSalaDeJuego(sala.Create_Clone);
  res:= uCosaConNombre.existeReferencia_al(cosaConNombre);
  salaAux.Free;
  uCosaConNombre.LimpiarReferencias;
  result:= res;
end;

procedure AlInicio;
begin
//	registrarClaseDeCosa( TSalaDeJuegoParaEditor.ClassName, TSalaDeJuegoParaEditor, true );
end;

procedure AlFinal;
begin
end;

end.

