unit uPostOpers;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

uses
{$IFDEF WINDOWS}
  Dialogs,
{$ENDIF}
  uHistoVarsOps, xMatDefs, Math, uSimResGlobs, uAuxiliares, SysUtils;

type

  TClaseDePostOper = class of TPostOper;

  TPostOper = class
  public
    res: TCronVar;
    //Resuelve los valores que tengan que ver con los ndices o el
    //encabezado de la simulacion.
    //Para las PostOpers debe inicializar los resultados de manera que tengan
    //las dimensiones necesarias para alojar sus resultados, p ej en una suma
    //de CronVars las mismas dimensiones que los parametros, en un
    //cambioPasoDeTiempo la cantidad de cnicas del parmetro y la cantidad de
    //pasos del parametro * (viejaDurPaso / nuevaDurPaso)
    procedure prepararse; virtual; abstract;
    procedure Evaluar; virtual; abstract;
    class function tipo: string; virtual; abstract;
    function referenciaCronVar(cronVar: TCronVar): boolean; virtual;
    function parametrosCronVar: TDAOfCronVar; virtual; abstract;
    function nombresParametrosCronVar: string; virtual; abstract;
    function parametrosAdicionales: string; virtual;
    function serialize: string; virtual;
  end;

  TPostOperUnaCronVar = class(TPostOper)
  public
    param1: TCronVar;
    function referenciaCronVar(cronVar: TCronVar): boolean; override;
    function parametrosCronVar: TDAOfCronVar; override;
    function nombresParametrosCronVar: string; override;
    function serialize: string; override;
  end;

  TPostOperUnaCronVarUnReal = class(TPostOperUnaCronVar)
  public
    real: NReal;
    function serialize: string; override;
  end;

  TPostOperDosCronVars = class(TPostOperUnaCronVar)
  public
    param2: TCronVar;
    function referenciaCronVar(cronVar: TCronVar): boolean; override;
    function parametrosCronVar: TDAOfCronVar; override;
    function nombresParametrosCronVar: string; override;
    function serialize: string; override;
  end;

  TPostOperMultiCronVar = class(TPostOper)
  public
    params: TDAOfCronVar;
    function referenciaCronVar(cronVar: TCronVar): boolean; override;
    function parametrosCronVar: TDAOfCronVar; override;
    function nombresParametrosCronVar: string; override;
  end;


  TPostOper_minEntreCronVarYReal = class(TPostOperUnaCronVarUnReal)
  public
    constructor Create(res, param1: TCronVar; real: NReal);
    procedure prepararse; override;
    procedure Evaluar; override;
    class function tipo: string; override;
    function parametrosAdicionales: string; override;
  end;

  TPostOper_maxEntreCronVarYReal = class(TPostOperUnaCronVarUnReal)
  public
    constructor Create(res, param1: TCronVar; real: NReal);
    procedure prepararse; override;
    procedure Evaluar; override;
    class function tipo: string; override;
    function parametrosAdicionales: string; override;
  end;

  TPostOper_cronVarMasReal = class(TPostOperUnaCronVarUnReal)
  public
    constructor Create(res, param1: TCronVar; real: NReal);
    procedure prepararse; override;
    procedure Evaluar; override;
    class function tipo: string; override;
    function parametrosAdicionales: string; override;
  end;

  TPostOper_cronVarPorReal = class(TPostOperUnaCronVarUnReal)
  public
    constructor Create(res, param1: TCronVar; real: NReal);
    procedure prepararse; override;
    procedure Evaluar; override;
    class function tipo: string; override;
    function parametrosAdicionales: string; override;
  end;

  TPostOper_restaCronVars = class(TPostOperDosCronVars)
  public
    constructor Create(res, param1, param2: TCronVar);
    procedure prepararse; override;
    procedure Evaluar; override;
    class function tipo: string; override;
  end;

  (* Hace la combinacin lineal ponderarda de un conjunto de variables
  combinarCronVars resVar N coef1 coef2 .... coefN cronVar1 cronVar2 ... cronVarN
  *)
  TPostOper_combinarCronVars = class(TPostOperMultiCronVar)
  public
    coeficientes: TDAOfNReal;
    constructor Create(res: TCronVar; acronvars: TDAOfCronVar;
      coeficientes: TDAOfNReal);
    procedure prepararse; override;
    procedure Evaluar; override;
    class function tipo: string; override;
    function parametrosAdicionales: string; override;
    function serialize: string; override;
  end;

  (* Hace la combinacin lineal ponderarda de un conjunto de variables
  combinarCronVars resVar N coef1 coef2 .... coefN cronVar1 cronVar2 ... cronVarN
  Pero considerando cada una con un desplazamiento en el tiempo.
  Cuando al aplicar un desplazamiento, se sale de la serie, ese valor no se considera
  en las sumas por lo tanto, los bordes del resultado pueden no tener el aporte
  de una serie que al desplazarla no cubre el borde. Por ejemplo si res[k]= x[k]+y[k+1]
  res[0]= x[0]+y[1]
  ...
  res[k]= x[k]+y[k+1]
  ...
  res[n-1]=x[n-1]+y[n]
  res[n]= x[n] + 0

  como podemos ver en este caso, el ltimo valor de las series resultado est
  calculado con el aporte de slo una de las series.
  *)
  TPostOper_combinarDespCronVars = class(TPostOperMultiCronVar)
  public
    coeficientes:    TDAOfNReal;
    desplazamientos: TDAOfNInt;
    constructor Create(res: TCronVar; acronvars: TDAOfCronVar;
      coeficientes: TDAOfNReal; desplazamientos: TDAOfNInt);
    procedure prepararse; override;
    procedure Evaluar; override;
    class function tipo: string; override;
    function parametrosAdicionales: string; override;
    function serialize: string; override;
  end;

  (* res[iPaso][iCronica]:= q^iPaso * param1[iPaso][iCronica] *)
  TPostOper_aplicarActualizador = class(TPostOperUnaCronVarUnReal)
  public
    constructor Create(res, param1: TCronVar; q: NReal);
    procedure prepararse; override;
    procedure Evaluar; override;
    class function tipo: string; override;
    function parametrosAdicionales: string; override;
  end;

  TPostOper_multiplicacionCronVars = class(TPostOperDosCronVars)
  public
    constructor Create(res, param1, param2: TCronVar);
    procedure prepararse; override;
    procedure Evaluar; override;
    class function tipo: string; override;
  end;

  TPostOper_divisionCronVars = class(TPostOperDosCronVars)
  public
    constructor Create(res, param1, param2: TCronVar);
    procedure prepararse; override;
    procedure Evaluar; override;
    class function tipo: string; override;
  end;

  TPostOper_cambioPasoDeTiempo = class(TPostOperUnaCronVar)
  public
    horasPasoNuevo: NReal;
    promediar:      boolean;
    constructor Create(res, param1: TCronVar; horasPasoNuevo: NReal;
      promediar: boolean);
    procedure prepararse; override;
    procedure Evaluar; override;
    class function tipo: string; override;
    function parametrosAdicionales: string; override;
    function serialize: string; override;
  end;

  TPostOper_acumularCronVar = class(TPostOperUnaCronVar)
  public
    constructor Create(res, param1: TCronVar);
    procedure prepararse; override;
    procedure Evaluar; override;
    class function tipo: string; override;
  end;

  TPostOper_potenciaFirmeHidraulica = class(TPostOperMultiCronVar)
  public
    constructor Create(res: TCronVar; params: TDAOfCronVar);
    procedure prepararse; override;
    procedure Evaluar; override;
    class function tipo: string; override;
    procedure Free;
    function serialize: string; override;
  end;


  TPostOper_maximo = class(TPostOperMultiCronVar)
  public
    constructor Create(res: TCronVar; acronvars: TDAOfCronVar);
    procedure prepararse; override;
    procedure Evaluar; override;
    class function tipo: string; override;
    function serialize: string; override;
  end;

  TPostOper_MultiOrdenar = class(TPostOperMultiCronVar)
  public
    constructor Create(res: TCronVar; acronvars: TDAOfCronVar);
    procedure Evaluar; override;
    procedure prepararse; override;
    class function tipo: string; override;
    function serialize: string; override;
  end;

  TPostOper_MultiPromedioMovil = class(TPostOperMultiCronVar)
  public
    NPasosPM: integer;
    constructor Create(res: TCronVar; acronvars: TDAOfCronVar; NPasosPM: integer);
    procedure Evaluar; override;
    procedure prepararse; override;
    class function tipo: string; override;
    function serialize: string; override;
  end;


implementation

uses
  uAuxiliaresEscrituraSimRes;


 //--------------------
 //Mtodos de TPostOper
 //====================

function TPostOper.referenciaCronVar(cronVar: TCronVar): boolean;
begin
  Result := res = cronVar;
end;

function TPostOper.parametrosAdicionales: string;
begin
  Result := '-';
end;

function TPostOper.serialize: string;
begin
  Result := tipo + #9 + encomille(res.nombre);
end;

 //------------------------------
 //Mtodos de TPostOperUnaCronVar
 //==============================

function TPostOperUnaCronVar.referenciaCronVar(cronVar: TCronVar): boolean;
begin
  Result := inherited referenciaCronVar(cronVar) or (self.param1 = cronVar);
end;

function TPostOperUnaCronVar.parametrosCronVar: TDAOfCronVar;
var
  res: TDAOfCronVar;
begin
  SetLength(res, 1);
  res[0] := param1;
  Result := res;
end;

function TPostOperUnaCronVar.nombresParametrosCronVar: string;
begin
  Result := param1.nombre;
end;



function TPostOperUnaCronVar.serialize: string;
begin
  Result := inherited serialize + #9 + encomille(param1.nombre);
end;


 //------------------------------
 //Mtodos de TPostOperUnaCronVarUnReal
 //==============================
function TPostOperUnaCronVarUnReal.serialize: string;
begin
  Result := inherited serialize + #9 + FloatToStr(real);
end;


 //-------------------------------
 //Mtodos de TPostOperDosCronVars
 //===============================

function TPostOperDosCronVars.referenciaCronVar(cronVar: TCronVar): boolean;
begin
  Result := inherited referenciaCronVar(cronVar) or (self.param2 = cronVar);
end;

function TPostOperDosCronVars.parametrosCronVar: TDAOfCronVar;
var
  res: TDAOfCronVar;
begin
  SetLength(res, 2);
  res[0] := param1;
  res[1] := param2;
  Result := res;
end;

function TPostOperDosCronVars.nombresParametrosCronVar: string;
begin
  Result := inherited nombresParametrosCronVar + ', ' + param2.nombre;
end;

function TPostOperDosCronVars.serialize: string;
begin
  Result := inherited serialize + #9 + encomille(param2.nombre);
end;


 //--------------------------------
 //Mtodos de TPostOperMultiCronVar
 //================================

function TPostOperMultiCronVar.referenciaCronVar(cronVar: TCronVar): boolean;
var
  res: boolean;
  i:   integer;
begin
  res := inherited referenciaCronVar(cronVar);
  if not res then
  begin
    for i := 0 to high(params) do
      if cronVar = params[i] then
      begin
        res := True;
        break;
      end;
  end;
  Result := res;
end;

function TPostOperMultiCronVar.parametrosCronVar: TDAOfCronVar;
begin
  Result := params;
end;

function TPostOperMultiCronVar.nombresParametrosCronVar: string;
var
  res: string;
  i:   integer;
begin
  res := '';
  for i := 0 to high(params) - 1 do
    res  := res + params[i].nombre + ', ';
  res    := res + params[high(params)].nombre;
  Result := res;
end;

 //-----------------------------------------
 //Mtodos de TPostOper_minEntreCronVarYReal
 //=========================================

constructor TPostOper_minEntreCronVarYReal.Create(res, param1: TCronVar; real: NReal);
begin
  inherited Create;
  self.res    := res;
  self.param1 := param1;
  self.real   := real;
end;

procedure TPostOper_minEntreCronVarYReal.prepararse;
begin
  res.inicializar(param1.nCronicas, param1.nPasos, param1.dtIni, param1.DurPaso_horas);
end;

procedure TPostOper_minEntreCronVarYReal.Evaluar;
var
  i, j: integer;
  filaResi, filaParam1i: TDAofNReal;
begin
  for i := 0 to high(res.v) do
  begin
    filaResi    := res.v[i];
    filaParam1i := param1.v[i];
    for j := 0 to high(filaResi) do
      if real < filaParam1i[j] then
        filaResi[j] := real
      else
        filaResi[j] := filaParam1i[j];
  end;
end;

class function TPostOper_minEntreCronVarYReal.tipo: string;
begin
  Result := 'minEntreCronVarYReal';
end;

function TPostOper_minEntreCronVarYReal.parametrosAdicionales: string;
begin
  Result := 'real= ' + FloatToStrF(real, ffGeneral, 16, 10);
end;


 //-----------------------------------------
 //Mtodos de TPostOper_maxEntreCronVarYReal
 //=========================================

constructor TPostOper_maxEntreCronVarYReal.Create(res, param1: TCronVar; real: NReal);
begin
  inherited Create;
  self.res    := res;
  self.param1 := param1;
  self.real   := real;
end;

procedure TPostOper_maxEntreCronVarYReal.prepararse;
begin
  res.inicializar(param1.nCronicas, param1.nPasos, param1.dtIni, param1.DurPaso_horas);
end;

procedure TPostOper_maxEntreCronVarYReal.Evaluar;
var
  i, j: integer;
  filaResi, filaParam1i: TDAofNReal;
begin
  for i := 0 to high(res.v) do
  begin
    filaResi    := res.v[i];
    filaParam1i := param1.v[i];
    for j := 0 to high(filaResi) do
      if real > filaParam1i[j] then
        filaResi[j] := real
      else
        filaResi[j] := filaParam1i[j];
  end;
end;

class function TPostOper_maxEntreCronVarYReal.tipo: string;
begin
  Result := 'maxEntreCronVarYReal';
end;

function TPostOper_maxEntreCronVarYReal.parametrosAdicionales: string;
begin
  Result := 'real= ' + FloatToStrF(real, ffGeneral, 16, 10);
end;


 //-----------------------------------
 //Mtodos de TPostOper_CronVarMasReal
 //===================================

constructor TPostOper_CronVarMasReal.Create(res, param1: TCronVar; real: NReal);
begin
  inherited Create;
  self.res    := res;
  self.param1 := param1;
  self.real   := real;
end;

procedure TPostOper_CronVarMasReal.prepararse;
begin
  res.inicializar(param1.nCronicas, param1.nPasos, param1.dtIni, param1.DurPaso_horas);
end;

procedure TPostOper_CronVarMasReal.Evaluar;
var
  i, j: integer;
  filaResi, filaParam1i: TDAofNReal;
begin
  for i := 0 to high(res.v) do
  begin
    filaResi    := res.v[i];
    filaParam1i := param1.v[i];
    for j := 0 to high(filaResi) do
      filaResi[j] := filaParam1i[j] + real;
  end;
end;

class function TPostOper_CronVarMasReal.tipo: string;
begin
  Result := 'cronVarMasReal';
end;

function TPostOper_CronVarMasReal.parametrosAdicionales: string;
begin
  Result := 'real= ' + FloatToStrF(real, ffGeneral, 16, 10);
end;


 //-----------------------------------
 //Mtodos de TPostOper_CronVarPorReal
 //===================================

constructor TPostOper_CronVarPorReal.Create(res, param1: TCronVar; real: NReal);
begin
  inherited Create;
  self.res    := res;
  self.param1 := param1;
  self.real   := real;
end;

procedure TPostOper_CronVarPorReal.prepararse;
begin
  res.inicializar(param1.nCronicas, param1.nPasos, param1.dtIni, param1.DurPaso_horas);
end;

procedure TPostOper_CronVarPorReal.Evaluar;
var
  i, j: integer;
  filaResi, filaParam1i: TDAofNReal;
begin
  for i := 0 to high(res.v) do
  begin
    filaResi    := res.v[i];
    filaParam1i := param1.v[i];
    for j := 0 to high(filaResi) do
      filaResi[j] := filaParam1i[j] * real;
  end;
end;

class function TPostOper_CronVarPorReal.tipo: string;
begin
  Result := 'cronVarPorReal';
end;

function TPostOper_CronVarPorReal.parametrosAdicionales: string;
begin
  Result := 'real= ' + FloatToStrF(real, ffGeneral, 16, 10);
end;


 //----------------------------------
 //Mtodos de TPostOper_RestaCronVars
 //==================================

constructor TPostOper_RestaCronVars.Create(res, param1, param2: TCronVar);
begin
  inherited Create;
  self.res    := res;
  self.param1 := param1;
  self.param2 := param2;
end;

procedure TPostOper_RestaCronVars.prepararse;
begin
  res.inicializar(param1.nCronicas, param1.nPasos, param1.dtIni, param1.DurPaso_horas);
end;

procedure TPostOper_RestaCronVars.Evaluar;
var
  i, j: integer;
  filaResi, filaParam1i, filaParam2i: TDAofNReal;
begin
  for i := 0 to high(res.v) do
  begin
    filaResi    := res.v[i];
    filaParam1i := param1.v[i];
    filaParam2i := param2.v[i];
    for j := 0 to high(filaResi) do
      filaResi[j] := filaParam1i[j] - filaParam2i[j];
  end;
end;

class function TPostOper_RestaCronVars.tipo: string;
begin
  Result := 'restaCronVars';
end;

 //-------------------------------------
 //Mtodos de TPostOper_combinarCronVars
 //=====================================

constructor TPostOper_combinarCronVars.Create(res: TCronVar;
  acronvars: TDAOfCronVar; coeficientes: TDAOfNReal);
begin
  self.params := acronvars;
  self.coeficientes := coeficientes;
  self.res    := res;
end;

procedure TPostOper_combinarCronVars.prepararse;
begin
  res.inicializar(params[0].nCronicas, params[0].nPasos,
    params[0].dtIni, params[0].DurPaso_horas);
end;

procedure TPostOper_combinarCronVars.Evaluar;
var
  iPaso, jCronica: integer;
  ivar: integer;
  a:    NReal;
  hPasos, hCronicas: integer;
begin
  hPasos    := high(res.v);
  hCronicas := high(res.v[0]);

  for ivar := 0 to high(self.params) do
    writeln('Cronvar:', params[ivar].nombre,
      ', nPasos: ', length(params[ivar].v),
      ', nCronicas: ', length(params[ivar].v[0]));

  for iPaso := 0 to hPasos do
  begin
    for jCronica := 0 to hCronicas do
    begin
      a := 0;
      for ivar := 0 to high(self.params) do
        a := a + params[ivar].v[iPaso][jCronica] * coeficientes[ivar];
      res.v[iPaso][jCronica] := a;
    end;
  end;
end;

class function TPostOper_combinarCronVars.tipo: string;
begin
  Result := 'combinarCronVars';
end;

function TPostOper_combinarCronVars.parametrosAdicionales: string;
begin
  Result := 'coeficientes= ' + TDAOfNRealToStringSinTamanio(coeficientes, 16, 10, ', ');
end;


function TPostOper_combinarCronVars.serialize: string;
begin
  Result := inherited serialize + #9 + TDAOfNRealToTabbedStringConTamanio(
    coeficientes, 8, 3) + #9 + TDAOfCronVarToString(params);

end;

 //-----------------------------------------
 //Mtodos de TPostOper_combinarDespCronVars
 //=========================================

constructor TPostOper_combinarDespCronVars.Create(res: TCronVar;
  acronvars: TDAOfCronVar; coeficientes: TDAOfNReal; desplazamientos: TDAOfNInt);
begin
  self.params := acronvars;
  self.coeficientes := coeficientes;
  self.desplazamientos := desplazamientos;
  self.res    := res;
end;

procedure TPostOper_combinarDespCronVars.prepararse;
begin
  res.inicializar(params[0].nCronicas, params[0].nPasos,
    params[0].dtIni, params[0].DurPaso_horas);
end;

procedure TPostOper_combinarDespCronVars.Evaluar;
var
  iPaso, jCronica, iPasoDesp: integer;
  ivar: integer;
  a:    NReal;
  hPasos, hCronicas: integer;
begin
  hPasos    := high(res.v);    // cantidad de pasos de tiempo
  hCronicas := high(res.v[0]); // cantidad de crnicas
  for iPaso := 0 to hPasos do
  begin
    for jCronica := 0 to hCronicas do
    begin
      a := 0;
      for ivar := 0 to high(self.params) do
      begin
        iPasoDesp := iPaso + desplazamientos[ivar];
        if (iPasoDesp >= 0) and (iPasoDesp <= hPasos) then
          a := a + params[ivar].v[iPasoDesp][jCronica] * coeficientes[ivar];
      end;
      res.v[iPaso][jCronica] := a;
    end;
  end;
end;

class function TPostOper_combinarDespCronVars.tipo: string;
begin
  Result := 'combinarDespCronVars';
end;

function TPostOper_combinarDespCronVars.parametrosAdicionales: string;
begin
  Result := 'coeficientes= ' + TDAOfNRealToStringSinTamanio(coeficientes, 16, 10, ', ') +
    ', desplazamientos= ' + TDAOfNIntToStringSinTamanio(desplazamientos, ', ');
end;

function TPostOper_combinarDespCronVars.serialize: string;
begin
  Result := inherited serialize + #9 + TDAOfNRealToTabbedStringConTamanio(
    coeficientes, 8, 3) + #9 + TDAOfNIntToTabbedString(desplazamientos) +
    #9 + TDAOfCronVarToString(params);
end;

 //----------------------------------------
 //Mtodos de TPostOper_aplicarActualizador
 //========================================

constructor TPostOper_aplicarActualizador.Create(res, param1: TCronVar; q: NReal);
begin
  inherited Create;
  self.res    := res;
  self.param1 := param1;
  self.real   := q;
end;

procedure TPostOper_aplicarActualizador.prepararse;
begin
  res.inicializar(param1.nCronicas, param1.nPasos, param1.dtIni, param1.DurPaso_horas);
end;

procedure TPostOper_aplicarActualizador.Evaluar;
var
  iPaso, jCronica: integer;
  qa: NReal;
  hPasos, hCronicas: integer;
begin
  hPasos := high(res.v);
  hCronicas := high(res.v[0]);
  qa := 1;
  for iPaso := 0 to hPasos do
  begin
    for jCronica := 0 to hCronicas do
      res.v[iPaso][jCronica] := qa * param1.v[iPaso][jCronica];
    qa := qa * real;
  end;
end;

class function TPostOper_aplicarActualizador.tipo: string;
begin
  Result := 'aplicarActualizador';
end;

function TPostOper_aplicarActualizador.parametrosAdicionales: string;
begin
  Result := 'q= ' + FloatToStrF(real, ffGeneral, 16, 10);
end;


 //-------------------------------------------
 //Mtodos de TPostOper_multiplicacionCronVars
 //===========================================

constructor TPostOper_multiplicacionCronVars.Create(res, param1, param2: TCronVar);
begin
  inherited Create;
  self.res    := res;
  self.param1 := param1;
  Self.param2 := param2;
end;

procedure TPostOper_multiplicacionCronVars.prepararse;
begin
  res.inicializar(param1.nCronicas, param1.nPasos, param1.dtIni, param1.DurPaso_horas);
end;

procedure TPostOper_multiplicacionCronVars.Evaluar;
var
  i, j: integer;
  filaIRes, filaIParam1, filaIParam2: TDAofNReal;
begin
  for i := 0 to res.nPasos - 1 do
  begin
    filaIRes    := res.v[i];
    filaIParam1 := param1.v[i];
    filaIParam2 := param2.v[i];

    for j := 0 to high(filaIRes) do
      filaIRes[j] := filaIParam1[j] * filaIParam2[j];
  end;
end;

class function TPostOper_multiplicacionCronVars.tipo: string;
begin
  Result := 'multiplicacionCronVars';
end;

 //-------------------------------------
 //Mtodos de TPostOper_DivisionCronVars
 //=====================================

constructor TPostOper_DivisionCronVars.Create(res, param1, param2: TCronVar);
begin
  inherited Create;
  self.res    := res;
  self.param1 := param1;
  self.param2 := param2;
end;

procedure TPostOper_DivisionCronVars.prepararse;
begin
  res.inicializar(param1.nCronicas, param1.nPasos, param1.dtIni, param1.DurPaso_horas);
end;

procedure TPostOper_DivisionCronVars.Evaluar;
var
  i, j: integer;
  filaResi, filaParam1i, filaParam2i: TDAofNReal;
begin
  for i := 0 to high(res.v) do
  begin
    filaResi    := res.v[i];
    filaParam1i := param1.v[i];
    filaParam2i := param2.v[i];
    for j := 0 to high(filaResi) do
      filaResi[j] := filaParam1i[j] / filaParam2i[j];
  end;
end;

class function TPostOper_DivisionCronVars.tipo: string;
begin
  Result := 'divisionCronVars';
end;

 //---------------------------------------
 //Mtodos de TPostOper_cambioPasoDeTiempo
 //=======================================

constructor TPostOper_cambioPasoDeTiempo.Create(res, param1: TCronVar;
  horasPasoNuevo: NReal; promediar: boolean);
begin
  inherited Create;
  self.res    := res;
  self.param1 := param1;
  self.horasPasoNuevo := horasPasoNuevo;
  self.promediar := promediar;
end;

procedure TPostOper_cambioPasoDeTiempo.prepararse;
var
  nPasosRes: integer;
begin
  nPasosRes := trunc((param1.DurPaso_horas / horasPasoNuevo) * param1.nPasos);
  if nPasosRes <= 0 then
    raise Exception.Create(
      'TPostOper_cambioPasoDeTiempo, el resultado tendra CERO datos. Revise los parmetros.'
      );
  res.inicializar(param1.nCronicas, nPasosRes, param1.dtIni, horasPasoNuevo);
end;

procedure TPostOper_cambioPasoDeTiempo.Evaluar;
var
  j1, j2:     integer;
  jr1, jr2:   NReal;
  iCronica, kPaso: integer;
  filaKRes, filaK2Res, filaKParam1: TDAofNReal;
  h1sobreh2:  NReal;
  h2sobreh1:  NReal;
  alfa, beta: NReal;
begin
  h1sobreh2 := param1.DurPaso_horas / horasPasoNuevo;
  h2sobreh1 := horasPasoNuevo / param1.DurPaso_horas;
  for kPaso := 0 to high(res.v) do
  begin
    filaKRes := res.v[kpaso];
    for iCronica := 0 to high(filaKRes) do
      filaKRes[iCronica] := 0;
  end;

  for kPaso := 0 to param1.nPasos - 1 do
  begin
    jr1 := kPaso * h1sobreh2;
    jr2 := (kPaso + 1) * h1sobreh2;

    j1 := Trunc(jr1);
    j2 := Trunc(jr2);
    filaKParam1 := param1.v[kPaso];
    if j1 > high(res.v) then
      break;
    filaKRes := res.v[j1];

    if j1 = j2 then
    begin
      for iCronica := 0 to param1.nCronicas - 1 do
        filaKRes[iCronica] := filaKRes[iCronica] + filaKParam1[iCronica];
    end
    else
    begin
      alfa      := (j2 - jr1) * h2sobreh1;
      beta      := 1 - alfa; //(jr2 - j2) * h2sobreh1;
      AsumaCero := 1e-10;
      if not ((alfa >= -AsumaCero) and (alfa <= (1 + AsumaCero)) and
        (beta >= -AsumaCero) and (beta <= (1 + AsumaCero))) then
        raise Exception.Create(
          'TPostOper_cambioPasoDeTiempo.Evaluar->Error, alfa o beta fuera de rango');


      if (j2 <= high(res.v)) then
      begin
        //Cuando entre por aca y termine el for la fila j1 del resultado va a estar
        //terminada de calcular
        filaK2Res := res.v[j2];
        for iCronica := 0 to param1.nCronicas - 1 do
        begin
          filaKRes[iCronica]  := filaKRes[iCronica] + filaKParam1[iCronica] * alfa;
          filaK2Res[iCronica] := filaK2Res[iCronica] + filaKParam1[iCronica] * beta;
        end;
      end
      else
        for iCronica := 0 to param1.nCronicas - 1 do
          filaKRes[iCronica] := filaKRes[iCronica] + filaKParam1[iCronica] * alfa;
    end;
  end;

  if promediar then
  begin
    for kPaso := 0 to res.nPasos - 1 do
    begin
      filaKRes := res.v[kPaso];
      for iCronica := 0 to res.nCronicas - 1 do
        filaKRes[iCronica] := filaKRes[iCronica] * h1sobreh2;
    end;
  end;
end;

class function TPostOper_cambioPasoDeTiempo.tipo: string;
begin
  Result := 'cambioPasoDeTiempo';
end;

function TPostOper_cambioPasoDeTiempo.parametrosAdicionales: string;
begin
  Result := 'horasPasoNuevo= ' + FloatToStrF(horasPasoNuevo, ffGeneral, 16, 10) +
    ', promediar= ' + boolToSiNo(promediar);
end;


function TPostOper_cambioPasoDeTiempo.serialize: string;
begin
  Result := inherited serialize + #9 + FloatToStr(horasPasoNuevo) +
    #9 + BoolToStr(promediar);
end;

 //------------------------------------
 //Mtodos de TPostOper_acumularCronVar
 //====================================

constructor TPostOper_acumularCronVar.Create(res, param1: TCronVar);
begin
  inherited Create;
  self.res    := res;
  self.param1 := param1;
end;

procedure TPostOper_acumularCronVar.prepararse;
begin
  res.inicializar(param1.nCronicas, param1.nPasos, param1.dtIni, param1.DurPaso_horas);
end;

procedure TPostOper_acumularCronVar.Evaluar;
var
  i, j: integer;
  filaImenos1Res, filaIRes, filaIParam1: TDAofNReal;
begin
  filaIRes    := res.v[0];
  filaIParam1 := param1.v[0];
  for j := 0 to param1.nCronicas - 1 do
    filaIRes[j] := filaIParam1[j];
  for i := 1 to param1.nPasos - 1 do
  begin
    filaImenos1Res := res.v[i - 1];
    filaIRes    := res.v[i];
    filaIParam1 := param1.v[i];
    for j := 0 to param1.nCronicas - 1 do
      filaIRes[j] := filaImenos1Res[j] + filaIParam1[j];
  end;
end;

class function TPostOper_acumularCronVar.tipo: string;
begin
  Result := 'acumularCronVar';
end;

 //--------------------------------------------
 //Mtodos de TPostOper_potenciaFirmeHidraulica
 //============================================

constructor TPostOper_potenciaFirmeHidraulica.Create(res: TCronVar;
  params: TDAOfCronVar);
begin
  inherited Create;
  self.res    := res;
  self.params := params;
end;

procedure TPostOper_potenciaFirmeHidraulica.prepararse;
begin
  //se asume que todos los parametros tienen la misma cantidad de pasos y crnicas,
  //fecha inicial y tiempo entre los pasos
  //El resultado tendra una columna con la potencia firme reconocida de cada generador
  //pasado como parametro y una mas con la suma de las potencias firmes reconocidas
  res.inicializar(
    length(params) + 1,
    params[0].nPasos, params[0].dtIni, params[0].DurPaso_horas);
end;

procedure TPostOper_potenciaFirmeHidraulica.Evaluar;
var
  iPaso, iCentral: integer;
  sumaPots, error, minError: NReal;

  PSum, auxcv: TCronVar;
  auxsum: NReal;
  iCron:  integer;
  j95:    integer;
  jMinError: integer;
  P95Sum: NReal;
  NCentrales: integer;
begin
  auxcv := params[0];
  NCentrales := length(params);
  PSum := TCronVar.Create('PSum');
  PSum.inicializar(auxcv.nCronicas, auxcv.nPasos, auxcv.dtIni, auxcv.DurPaso_horas);

  for iPaso := 0 to params[0].nPasos - 1 do
  begin
    // Cargamos en PSum la suma de las potencias de las centrales.
    for iCron := 0 to PSum.nCronicas - 1 do
    begin
      auxsum := 0;
      for iCentral := 0 to NCentrales - 1 do
      begin
        auxcv  := params[iCentral];
        auxsum := auxsum + auxcv.v[iPaso][iCron];
      end;
      PSum.v[iPaso][iCron] := auxsum;
    end;

    // ahora ordenamos las potencias de cada central en forma decreciente
    for iCentral := 0 to NCentrales - 1 do
    begin
      auxcv := params[iCentral];
      QuickSort_Decreciente(auxcv.v[iPaso]);
    end;
    // ordenamos la suma de potencias en forma decreciente.
    QuickSort_Decreciente(PSum.v[iPaso]);

    // registramos la posicin de la prob. de excedencia de 95% de la suma
    // de potencias y su valor.
    j95    := trunc(0.95 * (PSum.nCronicas - 1));
    P95Sum := PSum.v[iPaso][j95];

    // ahora buscamos la suma de potencias de las centrales que con igual prob.
    // de excedencia en cada una, diste lo menos posible de la P95Sum determinada
    // en el paso anterior. As jminError quedar con un ndice que indica la
    // probabilidad de excedencia que al ser considerada la potencia de cada
    // central por separado con esa probabilidad, la suma de dichas potencias
    // es la ms cercana a la potencia de probabilidad de excedenica de 95%
    // del conjunto de las centrales.
    minError  := P95Sum;
    jminError := -1;
    for iCron := 0 to PSum.nCronicas - 1 do
    begin
      sumaPots := 0;
      for iCentral := 0 to NCentrales - 1 do
        sumaPots := sumaPots + params[iCentral].v[iPaso][iCron];
      error      := abs(sumaPots - P95Sum);
      if (error < minError) then
      begin
        jminError := iCron;
        minError  := error;
      end;
    end;

(*
    if (jminError > j95) then
      raise Exception.Create('Violacin del principio de Brandino SumaP95 > P95Suma ');
    // exigirle al conjunto una probabilidad de excedencia dada es un problema ms
    // estricto que exigirle a cada central por separado la misma probabilidad de
    // excedencia.
  *)

    // ahora cargamos en los resultados, la potencia reconocida para cada
    // central.
    sumaPots := 0;
    for iCentral := 0 to NCentrales - 1 do
    begin
      res.v[iPaso][iCentral] := params[iCentral].v[iPaso][jMinError];
      sumaPots := sumaPots + res.v[iPaso][iCentral];
    end;
    //    res.v[iPaso][res.nCronicas - 1]:= sumaPots;
    // en la ltima columna del resultado cargamos la potencia firme del
    // conjunto de las centrales. (observar que no es sumaPots)
    res.v[iPaso][NCentrales] := P95Sum;
  end;
end;

class function TPostOper_potenciaFirmeHidraulica.tipo: string;
begin
  Result := 'potenciaFirmeHidraulica';
end;

procedure TPostOper_potenciaFirmeHidraulica.Free;
begin
  SetLength(params, 0);
  inherited Free;
end;

function TPostOper_potenciaFirmeHidraulica.serialize: string;
begin
  Result := inherited serialize + #9 + TDAOfCronVarToStringConTamanio(params);
end;

 //---------------------------
 //Mtodos de TPostOper_maximo
 //===========================

constructor TPostOper_maximo.Create(res: TCronVar; acronvars: TDAOfCronVar);
begin
  inherited Create;
  self.params := acronvars;
  self.res    := res;
end;

procedure TPostOper_maximo.prepararse;
begin
  res.inicializar(params[0].nCronicas, params[0].nPasos,
    params[0].dtIni, params[0].DurPaso_horas);
end;

procedure TPostOper_maximo.Evaluar;
var
  iPaso, jCronica: integer;
  ivar: integer;
  max:  NReal;
  hPasos, hCronicas: integer;
begin
  hPasos    := high(res.v);    // cantidad de pasos de tiempo
  hCronicas := high(res.v[0]); // cantidad de crnicas
  for iPaso := 0 to hPasos do
  begin
    for jCronica := 0 to hCronicas do
    begin
      max := -MaxNReal;
      for ivar := 0 to high(self.params) do
        if params[ivar].v[iPaso][jCronica] > max then
          max := params[ivar].v[iPaso][jCronica];
      res.v[iPaso][jCronica] := max;
    end;
  end;
end;

class function TPostOper_maximo.tipo: string;
begin
  Result := 'maximo';
end;


function TPostOper_maximo.serialize: string;
begin
  Result := inherited serialize + #9 + TDAOfCronVarToStringConTamanio(params);
end;


 //----------------------------------
 //Mtodos de TPostOper_MultiOrdenar
 //==================================

constructor TPostOper_MultiOrdenar.Create(res: TCronVar; acronvars: TDAOfCronVar);
begin
  inherited Create;
  self.params := acronvars;
  self.res    := res;
end;



procedure TPostOper_MultiOrdenar.Evaluar;
var
  iPaso, jCronica: integer;
  ivar:    integer;
  min:     NReal;
  hPasos, hCronicas: integer;
  k, kmin: integer;
begin
  hPasos    := high(res.v);    // cantidad de pasos de tiempo
  hCronicas := high(res.v[0]); // cantidad de crnicas
  for iPaso := 0 to hPasos do
  begin
    for jCronica := 0 to hCronicas do
    begin
      min  := res.v[iPaso][jCronica];
      kmin := jCronica;
      for k := jCronica + 1 to hCronicas do
        if res.v[iPaso][k] < min then
        begin
          kmin := k;
          min  := res.v[iPaso][k];
        end;
      if ( kmin <> jCronica ) then
      begin
        vswap(res.v[iPaso][jCronica], res.v[iPaso][kmin]);
        for ivar := 0 to high(self.params) do
          vswap(params[ivar].v[iPaso][jCronica], params[ivar].v[iPaso][kmin]);
      end;
    end;
  end;
end;

procedure TPostOper_MultiOrdenar.prepararse;
begin
// Se supone que la variable resultado es la Ordenadora
end;


class function TPostOper_MultiOrdenar.tipo: string;
begin
  Result := 'MultiOrdenar';
end;


function TPostOper_MultiOrdenar.serialize: string;
begin
  Result := inherited serialize + #9 + TDAOfCronVarToStringConTamanio(params);
end;




 //----------------------------------
 //Mtodos de TPostOper_MultiPromedioMovil
 //==================================

constructor TPostOper_MultiPromedioMovil.Create(res: TCronVar;
  acronvars: TDAOfCronVar; NPasosPM: integer);
begin
  inherited Create;
  self.params   := acronvars;
  self.res      := res;
  self.NPasosPM := NPasosPM;
end;


procedure TPostOper_MultiPromedioMovil.prepararse;
begin
  res.inicializar(params[0].nCronicas, params[0].nPasos,
    params[0].dtIni, params[0].DurPaso_horas);
end;


procedure TPostOper_MultiPromedioMovil.Evaluar;
var
  iPaso, jCronica: integer;
  ivar:    integer;
  aprom:   NReal;
  hPasos, hCronicas: integer;
  k, kmin: integer;
  jPaso:   integer;

begin
  hPasos    := high(res.v);    // cantidad de pasos de tiempo -1
  hCronicas := high(res.v[0]); // cantidad de crnicas -1
  for iPaso := 0 to hPasos do
  begin
    for jCronica := 0 to hCronicas do
    begin
      aprom := 0;
      jPaso := 0;
      while (((iPaso - jPaso) >= 0) and (jPaso < NPasosPM)) do
      begin
        for iVar := 0 to High(Params) do
          aprom := aprom + params[iVar].v[iPaso - jPaso][jCronica];
        Inc(jPaso);
      end;
      res.v[iPaso][jCronica] := aprom / NPasosPM / length(Params);
    end;
  end;
end;

class function TPostOper_MultiPromedioMovil.tipo: string;
begin
  Result := 'MultiPromedioMovil';
end;


function TPostOper_MultiPromedioMovil.serialize: string;
begin
  Result := inherited serialize + #9 + TDAOfCronVarToStringConTamanio(
    params) + #9 + FloatToStr(NPasosPM);
end;


end.

