{ Definiciones Corrientes para uso en matemticas }
unit xmatdefs;
{$MODE Delphi}

interface

uses
  SysUtils,
  uevaluador,
  Math;

resourcestring
  rs_viguales_RangeCheckError = 'viguales, no coinciden los largos.';


type

(* Con la siguiente definicin, ganamos ademas de precisin
velocidad cuando estemos usando un coprocesador numrico.
  Si se estan utilizando las rutinas de emulacin, se aumenta
la precisin pero obiamente la velocidaddecrece.
  !!!OJO: no utilizar el tipo Real mezclado con estas cosas,
  supongamos que salvamos un archivo un nmero real, pasa a ser
  una complicacin leerlo ya que podemos haber cambiado las
  opciones de compilacin.

  Los programas que esten pensados para uso general debern
poder reconocer con que estado se salv un archivo, para lo
cual se deber salvar tambin un indicativo.

  LO MS SANO ES DEJAR LAS OPCIONES DE COMPILACIN SIEMPRE
CON N+, E+  AUNQUE SE PIERDA VELOCIDAD CUANDO NO SE TENGA
COPROCESADOR.

  EL problema del aumento de tamao de los vectores de reales
es solo un corriemiento del lmite que ya teniamos cuando
usabamos reales solamente. La solucin al problema pasa por
el uso de espacios de memoria virtual, para lo que hemos
desarrollado los Vectores Gigantes Paginads (VGP) *)

  NReal = double;
  PNReal = ^NReal;
  NInt = integer;
  PNInt = ^NInt;

const
  MaxNReal = Math.MaxDouble;

type
  NComplex = record
    r, i: NReal
  end;
  PNComplex = ^NComplex;

const
  Complex_UNO: NComplex = (r: 1; i: 0);


type
  TDAofNCardinal = array of cardinal;
  TDAofNInt = array of integer;
  PDAofNInt = ^TDAofNInt;
  TDAofNReal = array of NReal;
  PDAOfNReal = ^TDAOfNReal;
  TDAOfDateTime = array of TDateTime;
  TDAofNComplex = array of NComplex;
  TDAofString = array of string;
  TDAOfBoolean = array of boolean;
  TDAOfPtr = array of Pointer;

  TMatOfBoolean = array of TDAOfBoolean;

  TMatOfNInt = array of TDAofNInt;
  TDAOfDAOfNInt = TMatOfNInt;

  TMatOfNReal = array of TDAofNReal;
  TDAOfDAofNReal = TMatOfNReal;

  TDAOfMatOfNReal = array of TMatOfNReal;
  TMatOfString = array of TDAofString;
  TDAofDAofString = TMatOfString;


type
  TFdeX = function(x: NReal): NReal;


type
  NEntero = NInt;
  PNEntero = ^NInt;

type
  TDAofI = array of NInt;
  TSetOfChar = set of char;

{ Redefinicion de los timpos estandar para que el compilador detecte
como un error el uso indebido de los mismos }
//  extended = boolean;
  real = boolean;


{ Las siguientes constantes son calculadas en el auto-arranque de
la unidad por lo que no es aconsejable utilizarlas en los procedimientos
de auto-arranque de otras unidades pues de hacerlo habr que tener
}
var
  AsumaCero: NReal; {EPSILON de la maquina en cuentas con NReal }
  DosPi: NReal;    { 2*Pi }


{ abs(x) < AsumaCero }
function EsCero(x: NReal): boolean; overload;
function EsCero(z: NComplex): boolean; overload;


{ Casi0:= Abs(x)< xCero.
  xCero debe ser un numero positivo }
function Casi0(x: NReal; xCero: NReal): boolean;

function SignoNR(x: NReal): integer;

function vsum(const a: TDAofNReal): NReal; overload;
function vsum(const a: TDAofNInt): integer; overload;
function vprom(const a: TDAofNReal): NReal; overload;
function vprom(const a: TDAofNInt): NReal; overload;


(*+doc
  result:= ( a[i] = b[i] ) para todo i
-doc*)
function viguales(const a, b: TDAOfNReal): boolean;


(*+doc
  result:= sumatoria( a[i]*b[i] )
-doc*)
function vsumaproducto(const a, b: TDAOfNReal): NReal; overload;
function vsumaproducto(const a, b: TDAOfNInt): NReal; overload;

(*+doc
  a[i]:= b[i] ; i:= 0... high(b)
-doc*)
procedure vcopy(var a: TDAOfNReal; const b: TDAOfNReal); overload;

(*+doc
  a[i]:= b[i] ; i:= 0... high(b)
-doc*)
procedure vcopy(var a: TDAofNInt; const b: TDAofNInt); overload;
(*+doc
  destino[ i+ jDesde]:= origen[ i + jDesde] ; i:= 0 ... N-1
-doc*)
procedure vcopyTramo(var destino: TDAofNReal; const origen: TDAofNReal;
  jDesde, N: integer);

(*+doc
  destino[ i+ jDesdeDestino]:= origen[ i + jDesdeOrigen] ; i:= 0 ... N-1
-doc*)


procedure vcopyTramoDesplazando(var destino: TDAofNReal; jDesdeDestino: integer;
  const origen: TDAofNReal; jDesdeOrigen: integer; N: integer);overload;

procedure vcopyTramoDesplazando(var destino: TDAofNInt; jDesdeDestino: integer;
  const origen: TDAofNInt; jDesdeOrigen: integer; N: integer);overload;


procedure vmultr(var a: TDAofNReal; r: NReal); overload;
procedure vmultr(var a: TDAofNInt; r: NReal); overload;

procedure vclear(var a: TDAofNInt); overload;
procedure vclear(var a: TDAofNReal); overload;

procedure vclear(var a: TDAofNInt; kIni, kFin: integer); overload;
procedure vclear(var a: TDAofNReal; kIni, kFin: integer); overload;

procedure vswap(var a, b: NReal); overload;
procedure vswap(var a, b: TDAOfNReal); overload;
procedure vswap(var a, b: TDAOfNInt); overload;
procedure vswap(var a, b: integer); overload;
procedure vswap(var a, b: TDAOfBoolean); overload;

function vmax(const a: TDAofNReal): NReal; overload;
function vmax(const a: TDAofNInt): integer; overload;

function vmin(const a: TDAofNReal): NReal; overload;


//Suma a acum los valores de source
//source y acum deben tener el mismo tamao
procedure vacum(var acum: TDAofNReal; const Source: TDAofNReal); overload;
//Suma a acum count valores de source desde jIni
procedure vacum(jIni: integer; var acum: TDAofNReal; const Source: TDAofNReal;
  Count: integer); overload;
//Suma a acum desede jIniAcum count valores de source desde jIniSource
procedure vacum(var acum: TDAofNReal; jIniAcum: integer; const Source: TDAofNReal;
  jIniSource: integer; Count: integer); overload;

//Suma a acum los valores de source ponderados por ponder
//source y acum deben tener el mismo tamao
procedure vacumPonderado(var acum: TDAofNReal; const Source: TDAofNReal;
  ponder: NReal); overload;
//Suma a acum count valores de source desde jIni multipicados por ponder
procedure vacumPonderado(jIni: integer; var acum: TDAofNReal;
  const Source: TDAofNReal; Count: integer; ponder: NReal); overload;
//Suma a acum desede jIniAcum count valores de source desde jIniSource multipicados por ponder
procedure vacumPonderado(var acum: TDAofNReal; jIniAcum: integer;
  const Source: TDAofNReal; jIniSource: integer; Count: integer;
  ponder: NReal); overload;

procedure liberarMatriz(var mat: TMatOfNReal);
function create_clone(mat: TMatOfNReal): TMatOfNReal;

function moduloCiclico(dividendo, divisor: integer): cardinal;
//Retorna el resto de la divisin entera de dividendo / divisor
function moduloRealEntero(dividendo: NReal; divisor: integer): NReal;

// redefino esta funcin de sysutils para hacerla pasar primero
// por un evaluador de expresiones.
// para que esto funcione al incluir esta unidad en la clausula uses
// de otra tener la precausion de ponerla despues de SysUtils (si es que aparece).
function StrToFloat(const s: string): NReal;



function DAOfNRealToStr_(const valor: TDAOfNReal; precision: integer;
  decimales: integer; sep: char): string;
function StrToDAOfNReal_(s: string; sep: char): TDAOfNReal;

function DAOfNIntToStr(const valor: TDAOfNInt; sep: char): string;
function StrToDAOfNInt(s: string; sep: char): TDAOfNInt;

{$IFDEF FPC}
// reescribo por bug en la original.
function FloatToStrF(Value: NReal; format: TFloatFormat;
  Precision, Digits: integer): string;
{$ENDIF}

// Procesa el string cadena byte a byte res = ( res xor cadena[k] ) rot_left_32 1 ; k = 1 to length( cadena )
procedure crc32_in_res(var res: integer; const cadena: string);


procedure StoreInFile_DAOfNReal( var f: file; var a: TDAOfNReal );
procedure LoadFromFile_DAOfNReal( var f: file; var a: TDAOfNReal );
procedure StoreInFile_DAOfNInt( var f: file; var a: TDAOfNInt );
procedure LoadFromFile_DAOfNInt( var f: file; var a: TDAOfNInt );

// Escribe en el archivo de texto el vector encabezando con NElementos:
// y escribiendo NElementosPorLinea. Como separador usa #9 (tabulador).
procedure printvect( var f: textfile; const a : TDAOfNReal; NElementosPorLinea: integer );


{
 y(u) = a * u^2 + b * u + c
y0 = y(0); y1 = y(1); y2 = y(2)
}
function interpolacion_parabolica_012( const y0, y1, y2: NReal; u: NReal ): NReal;

{
BUsca el valor k en el array y si lo encuentra retorna la posicin.
Si no lo encuentra retorna -1
}
function  kInArray( const v: TDAOfNInt; k: integer ): integer;

{
Primero se fija si k est en v, si est retorna la posicin
si no est lo agrega y retonra la posicin.
}
function addToArray( var v: TDAofNInt; k: integer ): integer;

{
Si k est en v lo quita y retorna la posicin de donde fue quitado.
Si k no est en retorna -1
}
function delFromArray( var v: TDAOfNInt; k: integer ): integer;



function boolToInt( a: boolean ): integer;
function IntToBool( i: integer ): boolean;

implementation

{$IFDEF FPC}
uses
  uparche_sysstr_inc_fpc;

{
BUsca el valor k en el array y si lo encuentra retorna la posicin.
Si no lo encuentra retorna -1
}
function  kInArray( const v: TDAOfNInt; k: integer ): integer;
var
  j: integer;
  buscando: boolean;
begin
  buscando:= true;
  for j:= 0 to high( v ) do
    if v[j] = k then
    begin
      buscando:= false;
      break;
    end;
  if buscando then
   result:= -1
  else
    result:= j;
end;

{
Primero se fija si k est en v, si est retorna la posicin
si no est lo agrega y retonra la posicin.
}
function addToArray( var v: TDAofNInt; k: integer ): integer;
var
 j: integer;
begin
  j:= kInArray( v, k );
  if j < 0 then
  begin
    setlength( v, length(v) + 1 );
    v[high( v ) ]:= k;
    result:= high( v );
  end
  else
    result:= j;
end;

{
Si k est en v lo quita y retorna la posicin de donde fue quitado.
Si k no est en retorna -1
}
function delFromArray( var v: TDAOfNInt; k: integer ): integer;
var
 j, h: integer;
 res: TDAOfNInt;
begin
  j:= kInArray( v, k );
  if j >= 0 then
  begin
    setlength(res, length( v ) - 1 );
    for h:= 0 to j-1 do
      res[h]:= v[h];
    for h:= j+1 to high( v ) do
      res[h-1]:= v[h];
    setlength( v , 0 );
    v:= res;
  end;
  result:= j;
end;

function interpolacion_parabolica_012( const y0, y1, y2: NReal; u: NReal ): NReal;
{
y = a u^2 + b u + c
con u = ( x - x0 ) / dx

y0 =   c
y1 = a  + b + c
y2 = 4 a + 2 b + c

2a - c = y2 - 2 y1
a = ( y2 - 2 * y1 + c ) / 2
b = y1 - a - c
}
var
  a, b, c: NREal;
begin
  c:= y0;
  a:= ( y2 - 2* y1 + c ) / 2;
  b:= y1 - a - c ;
  result:= ( a * u + b ) * u + c;
end;

function FloatToStrF(Value: NReal; format: TFloatFormat;
  Precision, Digits: integer): string;
begin
  Result := FloatToStrF_(Value, format, Precision, Digits);
end;

{$ENDIF}



procedure crc32_in_res(var res: integer; const cadena: string);
var
  k: integer;
  i: integer;
begin
  for k := 1 to length(cadena) do
  begin
    i := Ord(cadena[k]);
    res := res xor i;
    if (res and $80000000) = 0 then
      res := res shl 1
    else
      res := (res shl 1) or 1;
  end;
end;


function NextPal(var s: string; sep: char): string;
var
  k1, k2: integer;
  ts: string;
  separadores: TSetOfChar;
begin
  separadores := [sep, ' '];
  k1 := 1;
  while (k1 <= Length(s)) and (s[k1] in Separadores) do
    Inc(k1);
  k2 := k1;
  while (k2 <= Length(s)) and not (s[k2] in Separadores) do
    Inc(k2);
  ts := copy(s, k1, k2 - k1);
  Delete(s, 1, k2);
  Result := ts;
end;



function DAOfNRealToStr_(const valor: TDAOfNReal; precision: integer;
  decimales: integer; sep: char): string;
var
  k: integer;
  res: string;
  seps: shortstring;
begin
  res := '';
  if sep <> ' ' then
    seps := sep + ' '
  else
    seps := ' ';

  if (valor <> nil) and (length(valor) > 0) then
  begin
    res := FloatToStrF(valor[0], ffFixed, precision, decimales);
    for k := 1 to high(valor) do
      res := res + seps + FloatToStrF(valor[k], ffFixed, precision, decimales);
  end;
  Result := res;
end;


function StrToDAOfNReal_(s: string; sep: char): TDAOfNReal;
var
  res: TDAOfNReal;
  cnt: integer;
  maxcnt: integer;
  v: NReal;
  pal: string;

begin
  setlength(res, 100);
  maxcnt := 100;
  cnt := 0;
  while s <> '' do
  begin
    pal := nextpal(s, sep);
    if pal <> '' then
    begin
      v := StrToFloat(pal);
      if cnt = maxcnt then
      begin
        Inc(maxcnt, 10);
        setlength(res, maxcnt);
      end;
      res[cnt] := v;
      Inc(cnt);
    end;
  end;
  setlength(res, cnt);
  Result := res;
end;


function DAOfNIntToStr(const valor: TDAOfNInt; sep: char): string;
var
  k: integer;
  res: string;
  seps: shortstring;
begin
  res := '';

  if sep <> ' ' then
    seps := sep + ' '
  else
    seps := ' ';

  if length(valor) > 0 then
  begin
    res := IntToStr(valor[0]);
    for k := 1 to high(valor) do
      res := res + seps + IntToStr(valor[k]);
  end;
  Result := res;
end;

function StrToDAOfNInt(s: string; sep: char): TDAOfNInt;
var
  res: TDAOfNInt;
  cnt: integer;
  maxcnt: integer;
  v: NInt;
  pal: string;
begin
  setlength(res, 100);
  maxcnt := 100;
  cnt := 0;
  while s <> '' do
  begin
    pal := nextpal(s, sep);
    if pal <> '' then
    begin
      v := StrToInt(pal);
      if cnt = maxcnt then
      begin
        Inc(maxcnt, 10);
        setlength(res, maxcnt);
      end;
      res[cnt] := v;
      Inc(cnt);
    end;
  end;
  setlength(res, cnt);
  Result := res;
end;



procedure StoreInFile_DAOfNReal( var f: file; var a: TDAOfNReal );
var
  n: integer;
begin
  n:= length( a );
  blockwrite( f, n, sizeof( n ) );
  blockwrite( f, a[0], sizeof( NReal ) * n );
end;

procedure LoadFromFile_DAOfNReal( var f: file; var a: TDAOfNReal );
var
  n: integer;
begin
  blockread( f, n, sizeOf(n) );
  setlength( a, n );
  blockread( f, a[0], SizeOf(NReal )* n );
end;

procedure StoreInFile_DAOfNInt( var f: file; var a: TDAOfNInt );
var
  n: integer;
begin
  n:= length( a );
  blockwrite( f, n, sizeof( n ) );
  blockwrite( f, a[0], sizeof( NInt ) * n );
end;

procedure LoadFromFile_DAOfNInt( var f: file; var a: TDAOfNInt );
var
  n: integer;
begin
  blockread( f, n, sizeOf(n) );
  setlength( a, n );
  blockread( f, a[0], SizeOf(NInt )* n );
end;


procedure printvect( var f: textfile; const a : TDAOfNReal; NElementosPorLinea: integer );
var
  k, cnt: integer;
begin
  writeln( f, 'NElementos:',#9, length( a ) );
  cnt:= 0;
  for k:= 0 to high( a ) do
  begin
    if cnt > 0 then write( f, #9 );
    write( f, a[k] );
    inc( cnt );
    if cnt = NElementosPorLinea then
    begin
      cnt:= 0;
      writeln( f );
    end;
  end;
  if cnt > 0 then
    writeln( f );
end;

procedure vcopy(var a: TDAOfNReal; const b: TDAOfNReal);
var
  i: integer;
begin
  {$IFOPT R+}
  if length(a) <> length(b) then
    raise Exception.Create('vcopy, no coinciden los largos.');
  {$ENDIF}
  for i := 0 to high(b) do
    a[i] := b[i];
end;

procedure vcopy(var a: TDAofNInt; const b: TDAofNInt);
var
  i: integer;
begin
  {$IFOPT R+}
  if length(a) <> length(b) then
    raise Exception.Create('vcopy, no coinciden los largos.');
  {$ENDIF}
  for i := 0 to high(b) do
    a[i] := b[i];
end;

procedure vcopyTramo(var destino: TDAofNReal; const origen: TDAofNReal;
  jDesde, N: integer);
var
  i: integer;
begin
  {$IFOPT R+}
  if length(Destino) < (jDesde + N) then
    raise Exception.Create('vcopyTramo, (jDesde+N) = ' + IntToStr(
      jDesde + N) + ' > length(Destino) = ' + IntToStr(length(destino)));

  if length(Origen) < (jDesde + N) then
    raise Exception.Create('vcopyTramo, (jDesde+N) = ' + IntToStr(
      jDesde + N) + ' > length(Origen) = ' + IntToStr(length(destino)));

  {$ENDIF}

  for i := 0 to N - 1 do
    destino[i + jDesde] := origen[i + jDesde];
end;


procedure vcopyTramoDesplazando(var destino: TDAofNReal; jDesdeDestino: integer;
  const origen: TDAofNReal; jDesdeOrigen: integer; N: integer);

var
  i: integer;
begin

  {$IFOPT R+}
  if length(Destino) < (jDesdeDestino + N) then
    raise Exception.Create('vcopyTramo2, (jDesdeDestino+N) = ' +
      IntToStr(jDesdeDestino + N) + ' > length(Destino) = ' +
      IntToStr(length(destino)));

  if length(Origen) < (jDesdeOrigen + N) then
    raise Exception.Create('vcopyTramo, (jDesdeOrigen+N) = ' + IntToStr(
      jDesdeOrigen + N) + ' > length(Origen) = ' + IntToStr(length(destino)));

  {$ENDIF}

  for i := 0 to N - 1 do
    destino[i + jDesdeDestino] := origen[i + jDesdeOrigen];
end;

procedure vcopyTramoDesplazando(var destino: TDAofNInt; jDesdeDestino: integer;
  const origen: TDAofNInt; jDesdeOrigen: integer; N: integer);
var
  i: integer;
begin

  {$IFOPT R+}
  if length(Destino) < (jDesdeDestino + N) then
    raise Exception.Create('vcopyTramo2, (jDesdeDestino+N) = ' +
      IntToStr(jDesdeDestino + N) + ' > length(Destino) = ' +
      IntToStr(length(destino)));

  if length(Origen) < (jDesdeOrigen + N) then
    raise Exception.Create('vcopyTramo, (jDesdeOrigen+N) = ' + IntToStr(
      jDesdeOrigen + N) + ' > length(Origen) = ' + IntToStr(length(destino)));

  {$ENDIF}

  for i := 0 to N - 1 do
    destino[i + jDesdeDestino] := origen[i + jDesdeOrigen];
end;

function vsumaproducto(const a, b: TDAOfNReal): NReal;
var
  i: integer;
  acum: NReal;
begin
  acum := 0;
  {$IFOPT R+}
  if length(a) <> length(b) then
    raise Exception.Create('vsumaproducto, no coinciden los largos.');
  {$ENDIF}

  for i := 0 to high(a) do
    acum := acum + a[i] * b[i];
  Result := acum;
end;


function vsumaproducto(const a, b: TDAOfNInt): NReal;
var
  i: integer;
  acum: NReal;
begin
  acum := 0;
  {$IFOPT R+}
  if length(a) <> length(b) then
    raise Exception.Create('vsumaproducto, no coinciden los largos.');
  {$ENDIF}

  for i := 0 to high(a) do
    acum := acum + a[i] * b[i];
  Result := acum;
end;




function viguales(const a, b: TDAOfNReal): boolean;
var
  i: integer;
  res: boolean;
begin
  if length(a) <> length(b) then
  begin
    Result := False;
    exit;
  end;

  res := True;
  for i := 0 to high(a) do
    if a[i] <> b[i] then
    begin
      res := False;
      break;
    end;
  Result := res;
end;

function vmin(const a: TDAofNReal): NReal;
var
  i: integer;
  res: NReal;
begin
  res := a[0];
  for i := 1 to high(a) do
    if a[i] < res then
      res := a[i];
  Result := res;
end;


function vmax(const a: TDAofNReal): NReal;
var
  i: integer;
  res: NReal;
begin
  res := a[0];
  for i := 1 to high(a) do
    if a[i] > res then
      res := a[i];
  Result := res;
end;

function vmax(const a: TDAofNInt): integer; overload;
var
  i, res: integer;
begin
  res := a[0];
  for i := 1 to high(a) do
    if a[i] > res then
      res := a[i];
  Result := res;
end;

procedure vacum(var acum: TDAofNReal; const Source: TDAofNReal);
var
  i: integer;
begin
  for i := 0 to high(acum) do
    acum[i] := acum[i] + Source[i];
end;

procedure vacum(jIni: integer; var acum: TDAofNReal; const Source: TDAofNReal;
  Count: integer); overload;
var
  i: integer;
begin
  for i := 0 to Count - 1 do
    acum[jIni + i] := acum[jIni + i] + Source[jIni + i];
end;

procedure vacum(var acum: TDAofNReal; jIniAcum: integer; const Source: TDAofNReal;
  jIniSource: integer; Count: integer); overload;
var
  i: integer;
begin
  for i := 0 to Count - 1 do
    acum[jIniAcum + i] := acum[jIniAcum + i] + Source[jIniSource + i];
end;

procedure vacumPonderado(var acum: TDAofNReal; const Source: TDAofNReal;
  ponder: NReal); overload;
var
  i: integer;
begin
  for i := 0 to high(acum) do
    acum[i] := acum[i] + Source[i] * ponder;
end;

procedure vacumPonderado(jIni: integer; var acum: TDAofNReal;
  const Source: TDAofNReal; Count: integer; ponder: NReal); overload;
var
  i: integer;
begin
  for i := 0 to Count - 1 do
    acum[jIni + i] := acum[jIni + i] + Source[jIni + i] * ponder;
end;

procedure vacumPonderado(var acum: TDAofNReal; jIniAcum: integer;
  const Source: TDAofNReal; jIniSource: integer; Count: integer;
  ponder: NReal); overload;
var
  i: integer;
begin
  for i := 0 to Count - 1 do
    acum[jIniAcum + i] := acum[jIniAcum + i] + Source[jIniSource + i] * ponder;
end;

procedure liberarMatriz(var mat: TMatOfNReal);
var
  i: integer;
begin
  if mat <> nil then
  begin
    for i := 0 to high(mat) do
      SetLength(mat[i], 0);
    SetLength(mat, 0);
    mat := nil;
  end;
end;

function create_clone(mat: TMatOfNReal): TMatOfNReal;
var
  i: integer;
  res: TMatOfNReal;

begin
  setlength(res, length(mat));
  for i := 0 to high(mat) do
    res[i] := copy(mat[i]);
  Result := res;
end;


function moduloCiclico(dividendo, divisor: integer): cardinal;
var
  res: integer;
begin
  res := dividendo mod divisor;
  if res >= 0 then
    Result := res
  else
    Result := res + divisor;
end;

function moduloRealEntero(dividendo: NReal; divisor: integer): NReal;
begin
  Result := frac(dividendo / divisor) * divisor;
end;

function StrToFloat(const s: string): NReal;
begin
  Result := uevaluador.evalStrToFloat(s);
end;




procedure vswap(var a, b: integer);
var
  tv: integer;
begin
  tv := a;
  a := b;
  b := tv;
end;

procedure vswap(var a, b: NReal);
var
  tv: NReal;
begin
  tv := a;
  a := b;
  b := tv;
end;

procedure vswap(var a, b: TDAOfNReal);
var
  tv: TDAOfNReal;
begin
  tv := a;
  a := b;
  b := tv;
end;


procedure vswap(var a, b: TDAOfNInt);
var
  tv: TDAOfNInt;
begin
  tv := a;
  a := b;
  b := tv;
end;

procedure vswap(var a, b: TDAOfBoolean);
var
  tv: TDAOfBoolean;
begin
  tv := a;
  a := b;
  b := tv;
end;



procedure vclear(var a: TDAofNInt; kIni, kFin: integer);
var
  k: integer;
begin
  for k := kIni to kFin do
    a[k] := 0;
end;

procedure vclear(var a: TDAofNReal; kIni, kFin: integer);
var
  k: integer;
begin
  for k := kIni to kFin do
    a[k] := 0;
end;


procedure vclear(var a: TDAofNReal);
var
  k: integer;
begin
  for k := 0 to high(a) do
    a[k] := 0;
end;

function vsum(const a: TDAofNReal): NReal;
var
  res: NReal;
  k: integer;
begin
  res := 0;
  for k := 0 to high(a) do
    res := res + a[k];
  Result := res;
end;

function vprom(const a: TDAofNReal): NReal;
begin
  Result := vsum(a) / length(a);
end;


procedure vmultr(var a: TDAofNReal; r: NReal);
var
  k: integer;
begin
  for k := 0 to high(a) do
    a[k] := a[k] * r;
end;

procedure vmultr(var a: TDAofNInt; r: NReal);
var
  m: NReal;
  k: integer;
begin
  for k := 0 to high(a) do
  begin
    m := a[k] * r;
    a[k] := trunc(m + 0.5);
  end;
end;


procedure vclear(var a: TDAofNInt);
var
  k: integer;
begin
  for k := 0 to high(a) do
    a[k] := 0;

end;

function vsum(const a: TDAofNInt): integer;
var
  res: integer;
  k: integer;
begin
  res := 0;
  for k := 0 to high(a) do
    res := res + a[k];
  Result := res;
end;

function vprom(const a: TDAofNInt): NReal;
begin
  Result := vsum(a) / length(a);
end;


function SignoNR(x: NReal): integer;
begin
  if x < 0 then
    signoNR := -1
  else if x = 0 then
    signoNR := 0
  else
    signoNR := 1;
end;


function EsCero(x: NReal): boolean;
begin
  EsCero := Abs(x) < AsumaCero;
end;

function EsCero(z: NComplex): boolean;
begin
  Result := EsCero(z.r) and EsCero(z.i);
end;

function Casi0(x: NReal; xCero: NReal): boolean;
begin
  Casi0 := abs(x) < xCero;
end;


function calceps: NReal;
{calceps
  This function returns the machine EPSILON or floating point tolerance,
  the smallest positive real number such that 1.0 + EPSILON > 1.0.
  EPSILON is needed to set various tolerances for different algorithms.
  While it could be entered as a constant, I prefer to calculate it, since
  users tend to move software between machines without paying attention to
  the computing environment. Note that more complete routines exist.
}
var
  e, e0: NReal;
  i: integer;
begin {calculate machine epsilon}
  e0 := 1;
  i := 0;
  repeat
    e0 := e0 / 2;
    e := 1 + e0;
    i := i + 1;
  until (e = 1.0) or (i = 500000); {note safety check}
  e0 := e0 * 2;
  { Writeln('Machine EPSILON =',e0);}
  calceps := e0;
end; {calceps}



function boolToInt( a: boolean ): integer;
begin
  if a then
    result:= 1
  else
    result:= 0;
end;

function IntToBool( i: integer ): boolean;
begin
  result:= i <> 0;
end;

begin
  AsumaCero := CalcEps;
  DosPi := 2 * Pi;
end.
